"""Second differences relative magnitude delta methods."""

from .sdrm import compute_conditional_cs_sdrm
from .sdrmb import compute_conditional_cs_sdrmb
from .sdrmm import compute_conditional_cs_sdrmm

__all__ = [
    "compute_conditional_cs_sdrm",
    "compute_conditional_cs_sdrmb",
    "compute_conditional_cs_sdrmm",
]



"""Wrapper for outcome regression DiD estimators."""

from typing import Any, NamedTuple

import numpy as np

from .estimators.reg_did_panel import reg_did_panel
from .estimators.reg_did_rc import reg_did_rc
from .print import print_did_result
from .utils import preprocess_drdid


class ORDIDResult(NamedTuple):
    """Result from the outcome regression DiD estimator."""

    att: float
    se: float
    uci: float
    lci: float
    boots: np.ndarray | None
    att_inf_func: np.ndarray | None
    call_params: dict[str, Any]
    args: dict[str, Any]


ORDIDResult = print_did_result(ORDIDResult)


def ordid(
    data,
    y_col,
    time_col,
    treat_col,
    id_col=None,
    covariates_formula=None,
    panel=True,
    weights_col=None,
    boot=False,
    boot_type="weighted",
    n_boot=999,
    inf_func=False,
):
    r"""Wrap the outcome regression DiD estimators for the ATT.

    This function is a wrapper for outcome regression DiD estimators.
    It calls the appropriate estimator based on the panel argument and
    performs pre-processing for the data.

    Parameters
    ----------
    data : pd.DataFrame
        The input data containing outcome, time, unit ID, treatment,
        and optionally covariates and weights.
    y_col : str
        Name of the column containing the outcome variable.
    time_col : str
        Name of the column containing the time periods (must have exactly 2 periods).
    treat_col : str
        Name of the column containing the treatment group indicator.
        For panel data: time-invariant indicator (1 if ever treated, 0 if never treated).
        For repeated cross-sections: treatment status in the post-period.
    id_col : str | None, default None
        Name of the column containing the unit ID. Required if panel=True.
    covariates_formula : str | None, default None
        A formula for the covariates to include in the model.
        Should be of the form "~ X1 + X2" (intercept is always included).
        If None, equivalent to "~ 1" (intercept only).
    panel : bool, default True
        Whether the data is panel (True) or repeated cross-sections (False).
        Panel data should be in long format with each row representing
        a unit-time observation.
    weights_col : str | None, default None
        Name of the column containing sampling weights.
        If None, all observations have equal weight.
        Weights are normalized to have mean 1.
    boot : bool, default False
        Whether to compute bootstrap standard errors.
    boot_type : {"weighted", "multiplier"}, default "weighted"
        Type of bootstrap to perform (only relevant if boot=True).
    n_boot : int, default 999
        Number of bootstrap repetitions (only relevant if boot=True).
    inf_func : bool, default False
        Whether to return the influence function values.

    Returns
    -------
    ORDIDResult
        NamedTuple containing:

        - *att*: The OR DiD point estimate.
        - *se*: The OR DiD standard error.
        - *uci*: The upper bound of a 95% confidence interval.
        - *lci*: The lower bound of a 95% confidence interval.
        - *boots*: Bootstrap draws of the ATT if boot=True.
        - *att_inf_func*: Influence function values if inf_func=True.
        - *call_params*: Original function call parameters.
        - *args*: Arguments used in the estimation.

    Examples
    --------
    Estimate the average treatment effect on the treated (ATT) using outcome regression
    with panel data from a job training program. The outcome regression approach models
    the conditional expectation of the outcome given covariates.

    .. ipython::

        In [1]: import moderndid
           ...: from moderndid.data import load_nsw
           ...:
           ...: nsw_data = load_nsw()
           ...:
           ...: att_result = moderndid.ordid(
           ...:     data=nsw_data,
           ...:     y_col="re",
           ...:     time_col="year",
           ...:     treat_col="experimental",
           ...:     id_col="id",
           ...:     panel=True,
           ...:     covariates_formula="~ age + educ + black + married + nodegree + hisp + re74",
           ...: )

        In [2]: print(att_result)

    For more robust inference, we can use weighted-bootstrap standard errors with
    propensity score trimming to handle extreme weights.

    .. ipython::
        :okwarning:

        In [3]: att_result_rc_boot = moderndid.ordid(
           ...:     data=nsw_data,
           ...:     y_col="re",
           ...:     time_col="year",
           ...:     treat_col="experimental",
           ...:     id_col="id",
           ...:     panel=True,
           ...:     covariates_formula="~ age + educ + black + married + nodegree + hisp + re74",
           ...:     boot=True,
           ...: )

        In [4]: print(att_result_rc_boot)

    Notes
    -----
    The outcome regression DiD estimator is based on a linear regression model for
    the outcome conditional on covariates, time period, and treatment status. For
    panel data, it estimates the ATT by comparing the change in outcomes for treated
    units to the predicted change for control units based on their covariate values.

    This estimator assumes that the conditional expectation function is correctly
    specified and that the parallel trends assumption holds conditional on covariates.
    Unlike the doubly robust estimator, it is not robust to misspecification of the
    outcome regression model.

    See Also
    --------
    drdid : Doubly robust DiD estimator.
    ipwdid : Inverse propensity weighted DiD estimator.

    References
    ----------
    .. [1] Heckman, J., Ichimura, H., and Todd, P. (1997),
           "Matching as an Econometric Evaluation Estimator: Evidence from
           Evaluating a Job Training Programme", Review of Economic Studies,
           vol. 64(4), p. 605–654. https://doi.org/10.2307/2971733

    .. [2] Sant'Anna, P. H. C. and Zhao, J. (2020),
           "Doubly Robust Difference-in-Differences Estimators."
           Journal of Econometrics, Vol. 219 (1), pp. 101-122.
           https://doi.org/10.1016/j.jeconom.2020.06.003
    """
    if panel and id_col is None:
        raise ValueError("id_col must be provided when panel=True")

    call_params = {
        "y_col": y_col,
        "time_col": time_col,
        "treat_col": treat_col,
        "id_col": id_col,
        "covariates_formula": covariates_formula,
        "data_shape": data.shape,
        "panel": panel,
        "weights_col": weights_col,
        "boot": boot,
        "boot_type": boot_type,
        "n_boot": n_boot,
        "inf_func": inf_func,
    }

    dp = preprocess_drdid(
        data=data,
        y_col=y_col,
        time_col=time_col,
        id_col=id_col if panel else "dummy_id",  # Dummy ID for RC data
        treat_col=treat_col,
        covariates_formula=covariates_formula,
        panel=panel,
        weights_col=weights_col,
        boot=boot,
        boot_type=boot_type,
        n_boot=n_boot,
        inf_func=inf_func,
    )

    if panel:
        result = reg_did_panel(
            y1=dp["y1"],
            y0=dp["y0"],
            d=dp["D"],
            covariates=dp["covariates"],
            i_weights=dp["weights"],
            boot=boot,
            boot_type=boot_type,
            nboot=n_boot,
            influence_func=inf_func,
        )
    else:
        result = reg_did_rc(
            y=dp["y"],
            post=dp["post"],
            d=dp["D"],
            covariates=dp["covariates"],
            i_weights=dp["weights"],
            boot=boot,
            boot_type=boot_type,
            nboot=n_boot,
            influence_func=inf_func,
        )

    args = result.args.copy()
    args.update(
        {
            "panel": panel,
            "normalized": True,
            "boot": boot,
            "boot_type": boot_type,
            "nboot": n_boot,
            "type": "or",
        }
    )

    return ORDIDResult(
        att=result.att,
        se=result.se,
        uci=result.uci,
        lci=result.lci,
        boots=result.boots,
        att_inf_func=result.att_inf_func,
        call_params=call_params,
        args=args,
    )



"""Sensitivity analysis using the approach of Rambachan and Roth."""

from typing import NamedTuple, Protocol, runtime_checkable

import numpy as np
import pandas as pd

from .sensitivity import (
    OriginalCSResult,
    construct_original_cs,
    create_sensitivity_results_rm,
    create_sensitivity_results_sm,
)
from .utils import basis_vector


class HonestDiDResult(NamedTuple):
    """Result from honest_did analysis."""

    robust_ci: pd.DataFrame
    original_ci: OriginalCSResult
    sensitivity_type: str


@runtime_checkable
class EventStudyProtocol(Protocol):
    """Protocol for event study result objects."""

    aggregation_type: str
    influence_func: np.ndarray | None
    event_times: np.ndarray | None
    att_by_event: np.ndarray | None
    estimation_params: dict


def honest_did(
    event_study,
    event_time=0,
    sensitivity_type="smoothness",
    grid_points=100,
    **kwargs,
):
    """Compute sensitivity analysis for event study estimates.

    Implements the approach of Rambachan and Roth (2021) for robust inference
    in difference-in-differences and event study designs.

    Parameters
    ----------
    event_study : AGGTEResult or similar
        Event study result object containing influence functions and estimates.
    event_time : int, default=0
        Event time to compute sensitivity analysis for. Default is 0 (on impact).
    sensitivity_type : {'smoothness', 'relative_magnitude'}, default='smoothness'
        Type of sensitivity analysis:

        - 'smoothness': Allows violations of linear trends in pre-treatment periods
        - 'relative_magnitude': Based on relative magnitudes of deviations from parallel trends

    grid_points : int, default=100
        Number of grid points for underlying test inversion.
    **kwargs : Additional parameters
        Additional parameters passed to sensitivity analysis functions:

        - method : CI method ('FLCI', 'Conditional', 'C-F', 'C-LF')
        - m_vec : Vector of M values for smoothness bounds
        - m_bar_vec : Vector of Mbar values for relative magnitude bounds
        - monotonicity_direction : 'increasing' or 'decreasing'
        - bias_direction : 'positive' or 'negative'
        - alpha : Significance level (default 0.05)

    Returns
    -------
    HonestDiDResult
        NamedTuple containing:

        - robust_ci : DataFrame with sensitivity analysis results
        - original_ci : Original confidence interval
        - sensitivity_type : Type of analysis performed

    Examples
    --------
    The ``honest_did`` function performs sensitivity analysis on event study estimates to assess
    the robustness of results to violations of parallel trends. We demonstrate this below with a
    staggered treatment adoption design.

    First, we need to compute an event study. The function requires an event study
    object that contains the dynamic influence functions, which we obtain by first computing group-time
    effects and then aggregating them.

    .. ipython::
        :okwarning:

        In [1]: import numpy as np
           ...: from moderndid import att_gt, aggte, load_mpdta
           ...: from moderndid.didhonest import honest_did
           ...
           ...: df = load_mpdta()
           ...: gt_result = att_gt(
           ...:     data=df,
           ...:     yname="lemp",
           ...:     tname="year",
           ...:     gname="first.treat",
           ...:     idname="countyreal",
           ...:     est_method="dr",
           ...:     bstrap=False
           ...: )
           ...: es_result = aggte(gt_result, type="dynamic")
           ...: print(es_result)

    Now we can apply the honest DiD sensitivity analysis. We can examine how robust
    our estimated treatment effects are when we allow for potential violations of the parallel
    trends assumption. We'll analyze the on-impact effect (event_time=0) using smoothness
    restrictions, which bound how much pre-treatment trends can deviate.

    .. ipython::
        :okwarning:

        In [2]: hd_result = honest_did(
           ...:     event_study=es_result,
           ...:     event_time=0,
           ...:     sensitivity_type="smoothness",
           ...:     m_vec=[0.01, 0.02, 0.03]
           ...: )
           ...:
           ...: hd_result

    The output shows the original confidence interval (which assumes parallel trends hold exactly)
    and robust confidence intervals under different smoothness bounds. The ``m_vec`` parameter specifies
    different bounds on how much the linear trend can change between consecutive periods. Larger
    values of :math:`M` allow for bigger violations of parallel trends. If the robust CIs remain informative
    even under reasonable violations, this provides evidence that the results are credible.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2021). A more credible approach to
       parallel trends. Review of Economic Studies.
    """
    if hasattr(event_study, "aggregation_type") and isinstance(event_study, EventStudyProtocol):
        return _honest_did_aggte(
            event_study=event_study,
            event_time=event_time,
            sensitivity_type=sensitivity_type,
            grid_points=grid_points,
            **kwargs,
        )
    raise TypeError(
        f"honest_did not implemented for object of type {type(event_study).__name__}. "
        "Expected AGGTEResult or similar event study object."
    )


def _honest_did_aggte(
    event_study,
    event_time=0,
    sensitivity_type="smoothness",
    grid_points=100,
    **kwargs,
):
    """Implement sensitivity analysis for event study objects.

    Parameters
    ----------
    event_study : AGGTEResult
        Event study result from aggte function.
    event_time : int, default=0
        Event time to compute sensitivity analysis for. Default is 0 (on impact).
    sensitivity_type : str
        Type of sensitivity analysis:

        - 'smoothness': Allows violations of linear trends in pre-treatment periods
        - 'relative_magnitude': Based on relative magnitudes of deviations from parallel trends

    grid_points : int, default=100
        Number of grid points for underlying test inversion.
    **kwargs : Additional parameters
        Additional parameters passed to sensitivity analysis functions:

        - method : CI method ('FLCI', 'Conditional', 'C-F', 'C-LF')
        - m_vec : Vector of M values for smoothness bounds
        - m_bar_vec : Vector of Mbar values for relative magnitude bounds
        - monotonicity_direction : 'increasing' or 'decreasing'
        - bias_direction : 'positive' or 'negative'
        - alpha : Significance level (default 0.05)

    Returns
    -------
    HonestDiDResult
        NamedTuple with:

        - robust_ci : DataFrame with sensitivity analysis results
        - original_ci : Original confidence interval
        - sensitivity_type : Type of analysis performed
    """
    if event_study.aggregation_type != "dynamic":
        raise ValueError("honest_did requires an event study (dynamic aggregation).")

    if event_study.influence_func is None:
        raise ValueError("Event study must have influence functions computed.")

    influence_func = event_study.influence_func

    if influence_func.ndim != 2:
        raise ValueError(
            f"Expected 2D influence function matrix for dynamic aggregation, got shape {influence_func.shape}. "
        )

    event_times = event_study.event_times
    att_estimates = event_study.att_by_event

    reference_period = -1

    pre_periods = event_times[event_times < reference_period]
    post_periods = event_times[event_times > reference_period]

    if len(pre_periods) > 1 and not np.all(np.diff(pre_periods) == 1):
        raise ValueError(
            "honest_did expects consecutive pre-treatment periods. Please re-code your event study accordingly."
        )

    if len(post_periods) > 1 and not np.all(np.diff(post_periods) == 1):
        raise ValueError(
            "honest_did expects consecutive post-treatment periods. Please re-code your event study accordingly."
        )

    has_reference = reference_period in event_times
    if has_reference:
        ref_idx = np.where(event_times == reference_period)[0][0]
        event_times_no_ref = np.delete(event_times, ref_idx)
        att_no_ref = np.delete(att_estimates, ref_idx)
        influence_func_no_ref = np.delete(influence_func, ref_idx, axis=1)
    else:
        event_times_no_ref = event_times
        att_no_ref = att_estimates
        influence_func_no_ref = influence_func

    n = influence_func_no_ref.shape[0]
    vcov_matrix = influence_func_no_ref.T @ influence_func_no_ref / (n * n)

    num_pre_periods = np.sum(event_times_no_ref < reference_period)
    num_post_periods = len(event_times_no_ref) - num_pre_periods

    if num_pre_periods <= 0:
        raise ValueError("Not enough pre-treatment periods for honest_did.")
    if num_post_periods <= 0:
        raise ValueError("Not enough post-treatment periods for honest_did.")

    # Create weight vector for the requested event time
    # Note that event_time is relative to treatment, so we need to find its position in post-periods
    post_event_times = event_times_no_ref[num_pre_periods:]
    if event_time not in post_event_times:
        available_times = ", ".join(map(str, post_event_times))
        raise ValueError(
            f"Event time {event_time} not found in post-treatment periods. Available event times: {available_times}"
        )

    event_time_idx = np.where(post_event_times == event_time)[0][0] + 1
    l_vec = basis_vector(index=event_time_idx, size=num_post_periods)

    original_ci = construct_original_cs(
        betahat=att_no_ref,
        sigma=vcov_matrix,
        num_pre_periods=num_pre_periods,
        num_post_periods=num_post_periods,
        l_vec=l_vec,
        alpha=kwargs.get("alpha", 0.05),
    )

    if sensitivity_type == "relative_magnitude":
        robust_ci = create_sensitivity_results_rm(
            betahat=att_no_ref,
            sigma=vcov_matrix,
            num_pre_periods=num_pre_periods,
            num_post_periods=num_post_periods,
            l_vec=l_vec,
            grid_points=grid_points,
            **kwargs,
        )
    elif sensitivity_type == "smoothness":
        robust_ci = create_sensitivity_results_sm(
            betahat=att_no_ref,
            sigma=vcov_matrix,
            num_pre_periods=num_pre_periods,
            num_post_periods=num_post_periods,
            l_vec=l_vec,
            grid_points=grid_points,
            **kwargs,
        )
    else:
        raise ValueError(f"sensitivity_type must be 'smoothness' or 'relative_magnitude', got {sensitivity_type}")

    return HonestDiDResult(
        robust_ci=robust_ci,
        original_ci=original_ci,
        sensitivity_type=sensitivity_type,
    )



"""Functions for inference under second differences with monotonicity restrictions."""

from typing import NamedTuple

import numpy as np
import scipy.optimize as opt

from ...arp_no_nuisance import compute_arp_ci
from ...arp_nuisance import compute_arp_nuisance_ci, compute_least_favorable_cv
from ...bounds import create_monotonicity_constraint_matrix
from ...delta.sd.sd import _create_sd_constraint_matrix, _create_sd_constraint_vector
from ...fixed_length_ci import compute_flci
from ...numba import find_rows_with_post_period_values
from ...utils import basis_vector


class DeltaSDMResult(NamedTuple):
    """Result from second differences with monotonicity identified set computation.

    Attributes
    ----------
    id_lb : float
        Lower bound of the identified set.
    id_ub : float
        Upper bound of the identified set.
    """

    id_lb: float
    id_ub: float


def compute_conditional_cs_sdm(
    betahat,
    sigma,
    num_pre_periods,
    num_post_periods,
    l_vec=None,
    m_bar=0,
    alpha=0.05,
    monotonicity_direction="increasing",
    hybrid_flag="FLCI",
    hybrid_kappa=None,
    post_period_moments_only=True,
    grid_points=1000,
    grid_lb=None,
    grid_ub=None,
    seed=None,
):
    r"""Compute conditional confidence set for :math:`\Delta^{SDI}(M)`.

    Computes a confidence set for :math:`l'\tau_{post}` that is valid conditional on the event study
    coefficients being in the identified set under the second differences with monotonicity
    restriction :math:`\Delta^{SDI}(M)`.

    The combined smoothness and monotonicity restriction, denoted :math:`\Delta^{SDI}(M)` in [2]_, is
    defined as the intersection of :math:`\Delta^{SD}(M)` and a monotonicity restriction :math:`\Delta^I`

    .. math::

        \Delta^{SDI}(M) := \Delta^{SD}(M) \cap \Delta^{I},

    where

    .. math::

        \Delta^{SD}(M) := \{\delta: |(\delta_{t+1} - \delta_t) - (\delta_t - \delta_{t-1})| \le M, \forall t\},

    and :math:`\Delta^{I} := \{\delta: \delta_t \ge \delta_{t-1}, \forall t\}` for an increasing trend. For a
    decreasing trend, the restriction is :math:`\Delta^{SD}(M) \cap (-\Delta^{I})`.

    This restriction is useful when economic theory suggests both smooth evolution of
    trends and monotonic behavior (e.g., secular trends expected to continue post-treatment).
    The intersection typically leads to smaller identified sets than using either restriction alone.

    Parameters
    ----------
    betahat : ndarray
        Estimated event study coefficients.
    sigma : ndarray
        Covariance matrix of betahat.
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    l_vec : ndarray, optional
        Vector defining parameter of interest. If None, defaults to first post-period.
    m_bar : float, default=0
        Smoothness parameter M for :math:`\Delta^{SDI}(M)`.
    alpha : float, default=0.05
        Significance level.
    monotonicity_direction : {'increasing', 'decreasing'}, default='increasing'
        Direction of monotonicity restriction.
    hybrid_flag : {'FLCI', 'LF', 'ARP'}, default='FLCI'
        Type of hybrid test.
    hybrid_kappa : float, optional
        First-stage size for hybrid test. If None, defaults to alpha/10.
    post_period_moments_only : bool, default=True
        If True, use only post-period moments for ARP test.
    grid_points : int, default=1000
        Number of grid points for confidence interval search.
    grid_lb : float, optional
        Lower bound for grid search.
    grid_ub : float, optional
        Upper bound for grid search.
    seed : int, optional
        Random seed for reproducibility.

    Returns
    -------
    dict or float
        Returns dict with 'grid' and 'accept' arrays.

    Notes
    -----
    :math:`\Delta^{SDI}(M)` is a polyhedron formed by the intersection of smoothness and monotonicity
    constraints. The confidence set is constructed using either FLCIs or the moment inequality
    approach from Section 3 of [2]_.

    As noted in [2]_, monotonicity restrictions are often motivated by economic arguments. For
    example, Lovenheim & Willen (2019) argue that pre-treatment trends in the "wrong direction"
    (opposite to treatment effects) support their findings.

    Unlike :math:`\Delta^{SD}(M)` alone, the optimal FLCI for :math:`\Delta^{SDI}(M)` has the same
    worst-case bias as for :math:`\Delta^{SD}(M)`, meaning FLCIs do not adapt to the additional
    monotonicity restriction.

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2021). Inference for linear
        conditional moment inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    if l_vec is None:
        l_vec = basis_vector(1, num_post_periods)
    if hybrid_kappa is None:
        hybrid_kappa = alpha / 10

    A_sdm = _create_sdm_constraint_matrix(
        num_pre_periods, num_post_periods, monotonicity_direction, post_period_moments_only=False
    )
    d_sdm = _create_sdm_constraint_vector(num_pre_periods, num_post_periods, m_bar, post_period_moments_only=False)

    if post_period_moments_only and num_post_periods > 1:
        post_period_indices = list(range(num_pre_periods, num_pre_periods + num_post_periods))
        rows_for_arp = find_rows_with_post_period_values(A_sdm, post_period_indices)
    else:
        rows_for_arp = None

    if num_post_periods == 1:
        return _compute_cs_sdm_no_nuisance(
            betahat=betahat,
            sigma=sigma,
            num_pre_periods=num_pre_periods,
            num_post_periods=num_post_periods,
            A_sdm=A_sdm,
            d_sdm=d_sdm,
            l_vec=l_vec,
            m_bar=m_bar,
            alpha=alpha,
            hybrid_flag=hybrid_flag,
            hybrid_kappa=hybrid_kappa,
            monotonicity_direction=monotonicity_direction,
            grid_points=grid_points,
            grid_lb=grid_lb,
            grid_ub=grid_ub,
            seed=seed,
        )

    hybrid_list = {"hybrid_kappa": hybrid_kappa}

    if hybrid_flag == "FLCI":
        flci_result = compute_flci(
            beta_hat=betahat,
            sigma=sigma,
            smoothness_bound=m_bar,
            n_pre_periods=num_pre_periods,
            n_post_periods=num_post_periods,
            post_period_weights=l_vec,
            alpha=hybrid_kappa,
        )

        hybrid_list["flci_l"] = flci_result.optimal_vec
        hybrid_list["flci_halflength"] = flci_result.optimal_half_length

        try:
            vbar, _, _, _ = np.linalg.lstsq(A_sdm.T, flci_result.optimal_vec, rcond=None)
            hybrid_list["vbar"] = vbar
        except np.linalg.LinAlgError:
            hybrid_list["vbar"] = np.zeros(A_sdm.shape[0])

    if grid_lb is None or grid_ub is None:
        if hybrid_flag == "FLCI" and grid_lb is None:
            grid_lb = (flci_result.optimal_vec @ betahat) - flci_result.optimal_half_length
        if hybrid_flag == "FLCI" and grid_ub is None:
            grid_ub = (flci_result.optimal_vec @ betahat) + flci_result.optimal_half_length

        if grid_lb is None or grid_ub is None:
            id_set = compute_identified_set_sdm(
                m_bar=m_bar,
                true_beta=np.zeros(num_pre_periods + num_post_periods),
                l_vec=l_vec,
                num_pre_periods=num_pre_periods,
                num_post_periods=num_post_periods,
                monotonicity_direction=monotonicity_direction,
            )
            sd_theta = np.sqrt(l_vec.flatten() @ sigma[num_pre_periods:, num_pre_periods:] @ l_vec.flatten())
            if grid_lb is None:
                grid_lb = id_set.id_lb - 20 * sd_theta
            if grid_ub is None:
                grid_ub = id_set.id_ub + 20 * sd_theta

    result = compute_arp_nuisance_ci(
        betahat=betahat,
        sigma=sigma,
        l_vec=l_vec,
        a_matrix=A_sdm,
        d_vec=d_sdm,
        num_pre_periods=num_pre_periods,
        num_post_periods=num_post_periods,
        alpha=alpha,
        hybrid_flag=hybrid_flag,
        hybrid_list=hybrid_list,
        grid_lb=grid_lb,
        grid_ub=grid_ub,
        grid_points=grid_points,
        rows_for_arp=rows_for_arp,
    )

    return {"grid": result.accept_grid[:, 0], "accept": result.accept_grid[:, 1]}


def compute_identified_set_sdm(
    m_bar,
    true_beta,
    l_vec,
    num_pre_periods,
    num_post_periods,
    monotonicity_direction="increasing",
):
    r"""Compute identified set for :math:`\Delta^{SDI}(M)`.

    Computes the identified set for :math:`l'\tau_{post}` under the restriction that the underlying
    trend :math:`\delta` lies in :math:`\Delta^{SDI}(M)`, which combines second differences bounds
    with a monotonicity restriction.

    The identified set is an interval :math:`[\theta^{lb}, \theta^{ub}]` derived from Lemma 2.1 in [2]_.
    The bounds are given by

    .. math::

        \theta^{lb}(\beta, \Delta) := l'\beta_{post} - \max_{\delta} \{l'\delta_{post} :
        \delta \in \Delta, \delta_{pre} = \beta_{pre}\}

        \theta^{ub}(\beta, \Delta) := l'\beta_{post} - \min_{\delta} \{l'\delta_{post} :
        \delta \in \Delta, \delta_{pre} = \beta_{pre}\},

    where :math:`\Delta` is :math:`\Delta^{SDI}(M)`, the intersection of the smoothness and
    monotonicity restrictions.

    Parameters
    ----------
    m_bar : float
        Smoothness parameter M. Bounds the second differences:
        :math:`|\delta_{t-1} - 2\delta_t + \delta_{t+1}| \leq M`.
    true_beta : ndarray
        True coefficient values (pre and post periods).
    l_vec : ndarray
        Vector defining parameter of interest.
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    monotonicity_direction : {'increasing', 'decreasing'}, default='increasing'
        Direction of monotonicity restriction.

    Returns
    -------
    DeltaSDMResult
        Lower and upper bounds of the identified set.

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2021). Inference for linear
        conditional moment inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    f_delta = np.concatenate([np.zeros(num_pre_periods), l_vec.flatten()])

    A_sdm = _create_sdm_constraint_matrix(num_pre_periods, num_post_periods, monotonicity_direction)
    d_sdm = _create_sdm_constraint_vector(num_pre_periods, num_post_periods, m_bar)

    A_eq = np.hstack([np.eye(num_pre_periods), np.zeros((num_pre_periods, num_post_periods))])
    b_eq = true_beta[:num_pre_periods]

    bounds = [(None, None) for _ in range(num_pre_periods + num_post_periods)]

    result_max = opt.linprog(
        c=-f_delta,
        A_ub=A_sdm,
        b_ub=d_sdm,
        A_eq=A_eq,
        b_eq=b_eq,
        bounds=bounds,
        method="highs",
    )

    result_min = opt.linprog(
        c=f_delta,
        A_ub=A_sdm,
        b_ub=d_sdm,
        A_eq=A_eq,
        b_eq=b_eq,
        bounds=bounds,
        method="highs",
    )

    if not result_max.success or not result_min.success:
        observed_val = l_vec.flatten() @ true_beta[num_pre_periods:]
        return DeltaSDMResult(id_lb=observed_val, id_ub=observed_val)

    # Compute bounds of identified set
    # ID set = observed value ± bias
    observed = l_vec.flatten() @ true_beta[num_pre_periods:]
    id_ub = observed - result_min.fun
    id_lb = observed + result_max.fun

    return DeltaSDMResult(id_lb=id_lb, id_ub=id_ub)


def _create_sdm_constraint_matrix(
    num_pre_periods,
    num_post_periods,
    monotonicity_direction="increasing",
    post_period_moments_only=False,
):
    r"""Create constraint matrix for :math:`\Delta^{SDI}(M)`.

    Combines second differences (SD) and monotonicity (I) constraints into a single
    constraint matrix :math:`A` such that :math:`\delta \in \Delta^{SDI}(M)` can be written
    as :math:`A \delta \leq d`.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    monotonicity_direction : {'increasing', 'decreasing'}, default='increasing'
        Direction of monotonicity restriction.
    post_period_moments_only : bool, default=False
        If True, use only post-period moments.

    Returns
    -------
    ndarray
        Constraint matrix :math:`A`.
    """
    A_sd = _create_sd_constraint_matrix(
        num_pre_periods=num_pre_periods,
        num_post_periods=num_post_periods,
        post_period_moments_only=post_period_moments_only,
    )

    A_m = create_monotonicity_constraint_matrix(
        num_pre_periods=num_pre_periods,
        num_post_periods=num_post_periods,
        monotonicity_direction=monotonicity_direction,
        post_period_moments_only=post_period_moments_only,
    )

    A = np.vstack([A_sd, A_m])

    return A


def _create_sdm_constraint_vector(
    num_pre_periods,
    num_post_periods,
    m_bar,
    post_period_moments_only=False,
):
    r"""Create constraint vector for :math:`\Delta^{SDI}(M)`.

    Creates vector :math:`d` such that :math:`\delta \in \Delta^{SDI}(M)` can be written
    as :math:`A \delta \leq d`.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    m_bar : float
        Smoothness parameter M.
    post_period_moments_only : bool, default=False
        If True, use only post-period moments.

    Returns
    -------
    ndarray
        Constraint vector d.
    """
    A_sd = _create_sd_constraint_matrix(
        num_pre_periods=num_pre_periods,
        num_post_periods=num_post_periods,
        post_period_moments_only=post_period_moments_only,
    )

    d_sd = _create_sd_constraint_vector(A_sd, m_bar)

    # Monotonicity constraints have RHS of 0
    if post_period_moments_only:
        d_m = np.zeros(num_post_periods if num_post_periods > 1 else num_pre_periods + num_post_periods)
    else:
        d_m = np.zeros(num_pre_periods + num_post_periods)

    d = np.concatenate([d_sd, d_m])

    return d


def _compute_cs_sdm_no_nuisance(
    betahat,
    sigma,
    num_pre_periods,
    num_post_periods,
    A_sdm,
    d_sdm,
    l_vec,
    m_bar,
    alpha,
    hybrid_flag,
    hybrid_kappa,
    monotonicity_direction,
    grid_points,
    grid_lb,
    grid_ub,
    seed,
):
    """Compute confidence set for single post-period case (no nuisance parameters)."""
    hybrid_list = {"hybrid_kappa": hybrid_kappa}

    if hybrid_flag == "FLCI":
        flci_result = compute_flci(
            beta_hat=betahat,
            sigma=sigma,
            smoothness_bound=m_bar,
            n_pre_periods=num_pre_periods,
            n_post_periods=num_post_periods,
            post_period_weights=l_vec,
            alpha=hybrid_kappa,
        )

        # For single post-period, we need the full optimal_vec for compute_arp_ci
        hybrid_list["flci_l"] = flci_result.optimal_vec
        hybrid_list["flci_halflength"] = flci_result.optimal_half_length

        if grid_ub is None:
            grid_ub = flci_result.optimal_vec @ betahat + flci_result.optimal_half_length
        if grid_lb is None:
            grid_lb = flci_result.optimal_vec @ betahat - flci_result.optimal_half_length

    else:  # LF or ARP
        if hybrid_flag == "LF":
            lf_cv = compute_least_favorable_cv(
                x_t=None,
                sigma=A_sdm @ sigma @ A_sdm.T,
                hybrid_kappa=hybrid_kappa,
                seed=seed,
            )
            hybrid_list["lf_cv"] = lf_cv

        if grid_ub is None or grid_lb is None:
            id_set = compute_identified_set_sdm(
                m_bar=m_bar,
                true_beta=np.zeros(num_pre_periods + num_post_periods),
                l_vec=l_vec,
                num_pre_periods=num_pre_periods,
                num_post_periods=num_post_periods,
                monotonicity_direction=monotonicity_direction,
            )
            sd_theta = np.sqrt(l_vec.flatten() @ sigma[num_pre_periods:, num_pre_periods:] @ l_vec.flatten())
            if grid_ub is None:
                grid_ub = id_set.id_ub + 20 * sd_theta
            if grid_lb is None:
                grid_lb = id_set.id_lb - 20 * sd_theta

    arp_kwargs = {
        "beta_hat": betahat,
        "sigma": sigma,
        "A": A_sdm,
        "d": d_sdm,
        "n_pre_periods": num_pre_periods,
        "n_post_periods": num_post_periods,
        "alpha": alpha,
        "hybrid_flag": hybrid_flag,
        "hybrid_kappa": hybrid_kappa,
        "grid_lb": grid_lb,
        "grid_ub": grid_ub,
        "grid_points": grid_points,
    }

    if hybrid_flag == "FLCI":
        if "flci_l" in hybrid_list:
            arp_kwargs["flci_l"] = hybrid_list["flci_l"]
        if "flci_halflength" in hybrid_list:
            arp_kwargs["flci_halflength"] = hybrid_list["flci_halflength"]
    elif hybrid_flag == "LF":
        if "lf_cv" in hybrid_list:
            arp_kwargs["lf_cv"] = hybrid_list["lf_cv"]

    result = compute_arp_ci(**arp_kwargs)

    return {"grid": result.theta_grid, "accept": result.accept_grid}



"""Compute Aggregated Treatment Effect Parameters."""

from __future__ import annotations

import warnings

import numpy as np

from .aggte_obj import AGGTEResult
from .mboot import mboot


def compute_aggte(
    multi_period_result,
    aggregation_type="group",
    balance_e=None,
    min_e=-np.inf,
    max_e=np.inf,
    dropna=False,
    bootstrap=None,
    bootstrap_iterations=None,
    confidence_band=None,
    alpha=None,
    clustervars=None,
):
    """Compute aggregated treatment effect parameters.

    Aggregates group-time average treatment effects into different summary
    measures based on the specified aggregation type.

    Parameters
    ----------
    multi_period_result : MPResult
        Multi-period result object containing group-time ATTs and their
        influence functions.
    aggregation_type : {'simple', 'dynamic', 'group', 'calendar'}, default='group'
        Type of aggregation to perform:

        - 'simple': Simple weighted average of all post-treatment ATTs
        - 'dynamic': Event-study aggregation by relative time
        - 'group': Aggregation by treatment group
        - 'calendar': Aggregation by calendar time
    balance_e : int, optional
        For event studies, the relative time period to balance the sample on.
        This ensures a balanced panel with respect to event time.
    min_e : float, default=-np.inf
        Minimum event time to include in dynamic aggregation.
    max_e : float, default=np.inf
        Maximum event time to include in aggregation.
    dropna : bool, default=False
        Whether to remove NA values before aggregation.
    bootstrap : bool, optional
        Whether to use bootstrap inference. If None, uses the value from mp_result.
    bootstrap_iterations : int, optional
        Number of bootstrap iterations. If None, uses the value from mp_result.
    confidence_band : bool, optional
        Whether to compute uniform confidence bands. If None, uses value from mp_result.
    alpha : float, optional
        Significance level. If None, uses the value from mp_result.
    clustervars : list[str], optional
        Variables to cluster on. If None, uses the value from mp_result.

    Returns
    -------
    AGGTEResult
        Aggregated treatment effect results with overall ATT, standard errors,
        and potentially disaggregated effects by event time.

    References
    ----------
    .. [1] Callaway, B., & Sant'Anna, P. H. (2021). Difference-in-differences
           with multiple time periods. Journal of Econometrics.
    """
    groups = multi_period_result.groups
    times = multi_period_result.times
    att = multi_period_result.att_gt
    influence_function = multi_period_result.influence_func
    n_units = multi_period_result.n_units or len(influence_function)

    estimation_params = multi_period_result.estimation_params.copy()
    if bootstrap is not None:
        estimation_params["bootstrap"] = bootstrap
    if bootstrap_iterations is not None:
        estimation_params["biters"] = bootstrap_iterations
    if confidence_band is not None:
        estimation_params["uniform_bands"] = confidence_band
    if alpha is not None:
        estimation_params["alpha"] = alpha
    else:
        alpha = multi_period_result.alpha
    if clustervars is not None:
        estimation_params["clustervars"] = clustervars
    elif "clustervars" not in estimation_params:
        estimation_params["clustervars"] = None

    bootstrap = estimation_params.get("bootstrap", False)
    bootstrap_iterations = bootstrap_iterations or estimation_params.get("biters", 999)

    if aggregation_type not in ["simple", "dynamic", "group", "calendar"]:
        raise ValueError(
            f"`aggregation_type` must be one of ['simple', 'dynamic', 'group', 'calendar'], got {aggregation_type}"
        )

    if dropna:
        notna = ~np.isnan(att)
        groups = groups[notna]
        times = times[notna]
        att = att[notna]
        influence_function = influence_function[:, notna]

        if aggregation_type == "group":
            unique_groups = np.unique(groups)
            groups_with_att = []
            for g in unique_groups:
                mask = (groups == g) & (groups <= times)
                if mask.any() and not np.isnan(att[mask].mean()):
                    groups_with_att.append(g)

            keep = np.isin(groups, groups_with_att)
            groups = groups[keep]
            times = times[keep]
            att = att[keep]
            influence_function = influence_function[:, keep]

    elif np.any(np.isnan(att)):
        raise ValueError("Missing values at att_gt found. If you want to remove these, set `dropna = True'.")

    original_groups = groups.copy()
    original_times = times.copy()

    unique_original_groups = np.unique(original_groups)
    unique_original_times = np.unique(original_times)

    # Recode time periods to ensure they are 1 unit apart
    # This handles cases where time periods are not sequential integers
    unique_original_times_and_groups = np.sort(
        np.unique(np.concatenate([unique_original_times, unique_original_groups]))
    )
    recoded_times = np.arange(1, len(unique_original_times_and_groups) + 1)

    # Remap time-related variables to sequential integers
    groups = np.array([_orig2t(g, unique_original_times_and_groups, recoded_times) for g in original_groups])
    times = np.array([_orig2t(t, unique_original_times_and_groups, recoded_times) for t in original_times])
    unique_groups_recoded = np.array(
        [_orig2t(g, unique_original_times_and_groups, recoded_times) for g in unique_original_groups]
    )
    unique_times_recoded = np.unique(times)

    unit_level_groups = multi_period_result.G
    unit_level_weights = multi_period_result.weights_ind

    # If weights not provided, use uniform weights
    if unit_level_weights is None:
        if unit_level_groups is not None:
            unit_level_weights = np.ones(len(unit_level_groups))
        else:
            warnings.warn(
                "Unit-level data (unit_level_groups and unit_level_weights) not provided. "
                "Standard errors may be incorrect. Please update the MPResult object."
            )

            group_probabilities = np.array([np.mean(groups == g) for g in unique_groups_recoded])
            observation_group_probabilities = np.zeros(len(groups))
            for i, g in enumerate(unique_groups_recoded):
                mask = groups == g
                observation_group_probabilities[mask] = group_probabilities[i]
            unit_level_groups = None

    if unit_level_groups is not None and unit_level_weights is not None:
        unit_level_groups_recoded = np.array(
            [_orig2t(g, unique_original_times_and_groups, recoded_times) for g in unit_level_groups]
        )

        group_probabilities = np.array(
            [np.mean(unit_level_weights * (unit_level_groups_recoded == g)) for g in unique_groups_recoded]
        )
        observation_group_probabilities = np.zeros(len(groups))
        for i, g in enumerate(unique_groups_recoded):
            mask = groups == g
            observation_group_probabilities[mask] = group_probabilities[i]
    else:
        unit_level_groups_recoded = None

    post_treatment = (groups <= times) & (times <= groups + max_e)

    if aggregation_type == "simple":
        return _compute_simple_att(
            att=att,
            influence_function=influence_function,
            post_treatment=post_treatment,
            observation_group_probabilities=observation_group_probabilities,
            groups=groups,
            n_units=n_units,
            estimation_params=estimation_params,
            alpha=alpha,
            bootstrap_iterations=bootstrap_iterations,
            unique_groups_recoded=unique_groups_recoded,
            group_probabilities=group_probabilities,
            unit_level_groups=unit_level_groups_recoded,
            unit_level_weights=unit_level_weights,
        )

    if aggregation_type == "group":
        return _compute_group_att(
            att=att,
            influence_function=influence_function,
            groups=groups,
            times=times,
            unique_groups_recoded=unique_groups_recoded,
            unique_original_groups=unique_original_groups,
            group_probabilities=group_probabilities,
            n_units=n_units,
            max_e=max_e,
            estimation_params=estimation_params,
            alpha=alpha,
            bootstrap_iterations=bootstrap_iterations,
            unit_level_groups=unit_level_groups_recoded,
            unit_level_weights=unit_level_weights,
        )

    if aggregation_type == "dynamic":
        return _compute_dynamic_att(
            att=att,
            influence_function=influence_function,
            groups=groups,
            times=times,
            original_groups=original_groups,
            original_times=original_times,
            observation_group_probabilities=observation_group_probabilities,
            n_units=n_units,
            balance_e=balance_e,
            min_e=min_e,
            max_e=max_e,
            estimation_params=estimation_params,
            alpha=alpha,
            bootstrap_iterations=bootstrap_iterations,
            unique_groups_recoded=unique_groups_recoded,
            group_probabilities=group_probabilities,
            unit_level_groups=unit_level_groups_recoded,
            unit_level_weights=unit_level_weights,
            unique_original_times_and_groups=unique_original_times_and_groups,
            recoded_times=recoded_times,
        )

    if aggregation_type == "calendar":
        return _compute_calendar_att(
            att=att,
            influence_function=influence_function,
            groups=groups,
            times=times,
            unique_times_recoded=unique_times_recoded,
            observation_group_probabilities=observation_group_probabilities,
            n_units=n_units,
            estimation_params=estimation_params,
            alpha=alpha,
            bootstrap_iterations=bootstrap_iterations,
            unique_groups_recoded=unique_groups_recoded,
            group_probabilities=group_probabilities,
            unit_level_groups=unit_level_groups_recoded,
            unit_level_weights=unit_level_weights,
            unique_original_times_and_groups=unique_original_times_and_groups,
            recoded_times=recoded_times,
        )

    raise ValueError(f"Unexpected aggregation_type: {aggregation_type}")


def _compute_simple_att(
    att,
    influence_function,
    post_treatment,
    observation_group_probabilities,
    groups,
    n_units,
    estimation_params,
    alpha,
    bootstrap_iterations,
    unique_groups_recoded,
    group_probabilities,
    unit_level_groups,
    unit_level_weights,
):
    """Compute simple ATT by averaging all post-treatment effects."""
    # Simple ATT: weighted average of all post-treatment ATT(g,t)
    weights = observation_group_probabilities[post_treatment] / observation_group_probabilities[post_treatment].sum()
    simple_att = np.sum(att[post_treatment] * weights)

    if np.isnan(simple_att):
        simple_att = np.nan
        simple_se = np.nan
        simple_influence_function = np.zeros(n_units)
    else:
        # Get influence function for weights
        if unit_level_groups is not None and unit_level_weights is not None:
            weight_influence_function = _compute_weight_inf_func(
                keepers=np.where(post_treatment)[0],
                group_probabilities=group_probabilities,
                unit_level_weights=unit_level_weights,
                unit_level_groups=unit_level_groups,
                groups=groups,
                unique_groups_recoded=unique_groups_recoded,
            )
        else:
            weight_influence_function = None

        simple_influence_function = _get_agg_inf_func(
            att=att,
            influence_function=influence_function,
            keepers=post_treatment,
            weights=weights,
            weight_influence_function=weight_influence_function,
        )

        simple_se = _compute_se(
            influence_function=simple_influence_function,
            n_units=n_units,
            estimation_params=estimation_params,
            bootstrap_iterations=bootstrap_iterations,
            alpha=alpha,
        )

    return AGGTEResult(
        overall_att=simple_att,
        overall_se=simple_se,
        aggregation_type="simple",
        influence_func=simple_influence_function,
        influence_func_overall=simple_influence_function,
        estimation_params=estimation_params,
        call_info={"function": "compute_aggte", "type": "simple"},
    )


def _compute_group_att(
    att,
    influence_function,
    groups,
    times,
    unique_groups_recoded,
    unique_original_groups,
    group_probabilities,
    n_units,
    max_e,
    estimation_params,
    alpha,
    bootstrap_iterations,
    unit_level_groups,
    unit_level_weights,
):
    """Compute group-specific ATTs."""
    group_att = np.zeros(len(unique_groups_recoded))
    group_influence_functions = []

    for i, g in enumerate(unique_groups_recoded):
        mask = (groups == g) & ((g - 1) <= times) & (times <= g + max_e)

        if mask.any():
            group_att[i] = np.mean(att[mask])

            # Group-specific influence function
            weights_g = np.ones(mask.sum()) / mask.sum()
            influence_function_g = _get_agg_inf_func(
                att=att,
                influence_function=influence_function,
                keepers=mask,
                weights=weights_g,
                weight_influence_function=None,
            )
        else:
            group_att[i] = np.nan
            influence_function_g = np.zeros(n_units)

        group_influence_functions.append(influence_function_g)

    group_influence_functions = np.column_stack(group_influence_functions)
    group_att_clean = np.where(np.isnan(group_att), 0, group_att)

    group_se = np.zeros(len(unique_groups_recoded))
    for i in range(len(unique_groups_recoded)):
        if not np.isnan(group_att[i]):
            group_se[i] = _compute_se(
                influence_function=group_influence_functions[:, i],
                n_units=n_units,
                estimation_params=estimation_params,
                bootstrap_iterations=bootstrap_iterations,
                alpha=alpha,
            )
        else:
            group_se[i] = np.nan

    critical_value = _get_z_critical(alpha / 2)
    if estimation_params.get("uniform_bands", False):
        if not estimation_params.get("bootstrap", False):
            warnings.warn("Used bootstrap procedure to compute simultaneous confidence band")

        mboot_result = mboot(
            inf_func=group_influence_functions,
            n_units=n_units,
            biters=bootstrap_iterations,
            alp=alpha,
        )
        critical_value = mboot_result["crit_val"]

        if np.isnan(critical_value) or np.isinf(critical_value):
            warnings.warn(
                "Simultaneous critical value is NA. This probably happened because "
                "we cannot compute t-statistic (std errors are NA). "
                "We then report pointwise conf. intervals."
            )
            critical_value = _get_z_critical(alpha / 2)
            estimation_params["uniform_bands"] = False

    # Overall ATT: weighted average across groups
    overall_att = np.sum(group_att_clean * group_probabilities) / np.sum(group_probabilities)

    if unit_level_groups is not None and unit_level_weights is not None:
        weight_influence_function = _compute_weight_inf_func(
            keepers=np.arange(len(unique_groups_recoded)),
            group_probabilities=group_probabilities,
            unit_level_weights=unit_level_weights,
            unit_level_groups=unit_level_groups,
            groups=unique_groups_recoded,
            unique_groups_recoded=unique_groups_recoded,
        )
    else:
        weight_influence_function = None

    overall_influence_function = _get_agg_inf_func(
        att=group_att_clean,
        influence_function=group_influence_functions,
        keepers=np.ones(len(unique_groups_recoded), dtype=bool),
        weights=group_probabilities / group_probabilities.sum(),
        weight_influence_function=weight_influence_function,
    )

    overall_se = _compute_se(
        influence_function=overall_influence_function,
        n_units=n_units,
        estimation_params=estimation_params,
        bootstrap_iterations=bootstrap_iterations,
        alpha=alpha,
    )

    return AGGTEResult(
        overall_att=overall_att,
        overall_se=overall_se,
        aggregation_type="group",
        event_times=unique_original_groups,
        att_by_event=group_att,
        se_by_event=group_se,
        critical_values=np.full(len(unique_groups_recoded), critical_value),
        influence_func=group_influence_functions,
        influence_func_overall=overall_influence_function,
        estimation_params=estimation_params,
        call_info={"function": "compute_aggte", "type": "group"},
    )


def _compute_dynamic_att(
    att,
    influence_function,
    groups,
    times,
    original_groups,
    original_times,
    observation_group_probabilities,
    n_units,
    balance_e,
    min_e,
    max_e,
    estimation_params,
    alpha,
    bootstrap_iterations,
    unique_groups_recoded,
    group_probabilities,
    unit_level_groups,
    unit_level_weights,
    unique_original_times_and_groups,
    recoded_times,
):
    """Compute dynamic (event-study) treatment effects."""
    event_times = original_times - original_groups
    unique_event_times = np.unique(event_times)
    unique_event_times = unique_event_times[np.isfinite(unique_event_times)]
    unique_event_times = np.sort(unique_event_times)

    include_balanced = np.ones(len(groups), dtype=bool)
    if balance_e is not None:
        max_recoded_time = times.max()
        max_original_time = _t2orig(max_recoded_time, unique_original_times_and_groups, recoded_times)
        # Only keep observations where we can observe balance_e periods after treatment
        include_balanced = (max_original_time - original_groups) >= balance_e

        event_times_balanced = event_times[include_balanced]
        unique_event_times = np.unique(event_times_balanced)
        unique_event_times = unique_event_times[np.isfinite(unique_event_times)]
        unique_event_times = np.sort(unique_event_times)

        min_recoded_time = times.min()
        min_original_time = _t2orig(min_recoded_time, unique_original_times_and_groups, recoded_times)
        min_possible_e = balance_e - max_original_time + min_original_time
        unique_event_times = unique_event_times[
            (unique_event_times >= min_possible_e) & (unique_event_times <= balance_e)
        ]

    unique_event_times = unique_event_times[(unique_event_times >= min_e) & (unique_event_times <= max_e)]

    dynamic_att = np.zeros(len(unique_event_times))
    dynamic_influence_functions = []

    for i, e in enumerate(unique_event_times):
        mask = (event_times == e) & include_balanced

        if mask.any():
            pg_e = observation_group_probabilities[mask]
            weights_e = pg_e / pg_e.sum()

            dynamic_att[i] = np.sum(att[mask] * weights_e)

            if unit_level_groups is not None and unit_level_weights is not None:
                weight_influence_function_e = _compute_weight_inf_func(
                    keepers=np.where(mask)[0],
                    group_probabilities=group_probabilities,
                    unit_level_weights=unit_level_weights,
                    unit_level_groups=unit_level_groups,
                    groups=groups,
                    unique_groups_recoded=unique_groups_recoded,
                )
            else:
                weight_influence_function_e = None

            influence_function_e = _get_agg_inf_func(
                att=att,
                influence_function=influence_function,
                keepers=mask,
                weights=weights_e,
                weight_influence_function=weight_influence_function_e,
            )
        else:
            dynamic_att[i] = np.nan
            influence_function_e = np.zeros(n_units)

        dynamic_influence_functions.append(influence_function_e)

    dynamic_influence_functions = np.column_stack(dynamic_influence_functions)

    dynamic_se = np.zeros(len(unique_event_times))
    for i in range(len(unique_event_times)):
        if not np.isnan(dynamic_att[i]):
            dynamic_se[i] = _compute_se(
                influence_function=dynamic_influence_functions[:, i],
                n_units=n_units,
                estimation_params=estimation_params,
                bootstrap_iterations=bootstrap_iterations,
                alpha=alpha,
            )
        else:
            dynamic_se[i] = np.nan

    critical_value = _get_z_critical(alpha / 2)
    if estimation_params.get("uniform_bands", False):
        if not estimation_params.get("bootstrap", False):
            warnings.warn("Used bootstrap procedure to compute simultaneous confidence band")

        mboot_result = mboot(
            inf_func=dynamic_influence_functions,
            n_units=n_units,
            biters=bootstrap_iterations,
            alp=alpha,
        )
        critical_value = mboot_result["crit_val"]

        if np.isnan(critical_value) or np.isinf(critical_value):
            warnings.warn(
                "Simultaneous critical value is NA. This probably happened because "
                "we cannot compute t-statistic (std errors are NA). "
                "We then report pointwise conf. intervals."
            )
            critical_value = _get_z_critical(alpha / 2)
            estimation_params["uniform_bands"] = False

    # Overall ATT: average over non-negative event times
    post_mask = unique_event_times >= 0
    if post_mask.any():
        overall_att = np.nanmean(dynamic_att[post_mask])

        # Overall influence function
        overall_influence_function = _get_agg_inf_func(
            att=dynamic_att[post_mask],
            influence_function=dynamic_influence_functions[:, post_mask],
            keepers=np.ones(post_mask.sum(), dtype=bool),
            weights=np.ones(post_mask.sum()) / post_mask.sum(),
            weight_influence_function=None,
        )

        # Overall standard error
        overall_se = _compute_se(
            influence_function=overall_influence_function,
            n_units=n_units,
            estimation_params=estimation_params,
            bootstrap_iterations=bootstrap_iterations,
            alpha=alpha,
        )
    else:
        overall_att = np.nan
        overall_se = np.nan
        overall_influence_function = np.zeros(n_units)

    return AGGTEResult(
        overall_att=overall_att,
        overall_se=overall_se,
        aggregation_type="dynamic",
        event_times=unique_event_times,
        att_by_event=dynamic_att,
        se_by_event=dynamic_se,
        critical_values=np.full(len(unique_event_times), critical_value),
        influence_func=dynamic_influence_functions,
        influence_func_overall=overall_influence_function,
        min_event_time=int(min_e) if np.isfinite(min_e) else None,
        max_event_time=int(max_e) if np.isfinite(max_e) else None,
        balanced_event_threshold=balance_e,
        estimation_params=estimation_params,
        call_info={"function": "compute_aggte", "type": "dynamic"},
    )


def _compute_calendar_att(
    att,
    influence_function,
    groups,
    times,
    unique_times_recoded,
    observation_group_probabilities,
    n_units,
    estimation_params,
    alpha,
    bootstrap_iterations,
    unique_groups_recoded,
    group_probabilities,
    unit_level_groups,
    unit_level_weights,
    unique_original_times_and_groups,
    recoded_times,
):
    """Compute calendar time effects."""
    min_group = groups.min()
    calendar_times = unique_times_recoded[unique_times_recoded >= min_group]

    original_calendar_times = np.array(
        [_t2orig(t, unique_original_times_and_groups, recoded_times) for t in calendar_times]
    )

    calendar_att = np.zeros(len(calendar_times))
    calendar_influence_functions = []

    for i, t in enumerate(calendar_times):
        mask = (times == t) & (groups <= t)

        if mask.any():
            pg_t = observation_group_probabilities[mask]
            weights_t = pg_t / pg_t.sum()

            calendar_att[i] = np.sum(att[mask] * weights_t)

            if unit_level_groups is not None and unit_level_weights is not None:
                weight_influence_function_t = _compute_weight_inf_func(
                    keepers=np.where(mask)[0],
                    group_probabilities=group_probabilities,
                    unit_level_weights=unit_level_weights,
                    unit_level_groups=unit_level_groups,
                    groups=groups,
                    unique_groups_recoded=unique_groups_recoded,
                )
            else:
                weight_influence_function_t = None

            influence_function_t = _get_agg_inf_func(
                att=att,
                influence_function=influence_function,
                keepers=mask,
                weights=weights_t,
                weight_influence_function=weight_influence_function_t,
            )
        else:
            calendar_att[i] = np.nan
            influence_function_t = np.zeros(n_units)

        calendar_influence_functions.append(influence_function_t)

    calendar_influence_functions = np.column_stack(calendar_influence_functions)

    calendar_se = np.zeros(len(calendar_times))
    for i in range(len(calendar_times)):
        if not np.isnan(calendar_att[i]):
            calendar_se[i] = _compute_se(
                influence_function=calendar_influence_functions[:, i],
                n_units=n_units,
                estimation_params=estimation_params,
                bootstrap_iterations=bootstrap_iterations,
                alpha=alpha,
            )
        else:
            calendar_se[i] = np.nan

    critical_value = _get_z_critical(alpha / 2)
    if estimation_params.get("uniform_bands", False):
        if not estimation_params.get("bootstrap", False):
            warnings.warn("Used bootstrap procedure to compute simultaneous confidence band")

        mboot_result = mboot(
            inf_func=calendar_influence_functions,
            n_units=n_units,
            biters=bootstrap_iterations,
            alp=alpha,
        )
        critical_value = mboot_result["crit_val"]

        if np.isnan(critical_value) or np.isinf(critical_value):
            warnings.warn(
                "Simultaneous critical value is NA. This probably happened because "
                "we cannot compute t-statistic (std errors are NA). "
                "We then report pointwise conf. intervals."
            )
            critical_value = _get_z_critical(alpha / 2)
            estimation_params["uniform_bands"] = False

    # Overall ATT: simple average across time periods
    overall_att = np.nanmean(calendar_att)

    # Overall influence function
    n_valid = (~np.isnan(calendar_att)).sum()
    if n_valid > 0:
        overall_influence_function = _get_agg_inf_func(
            att=calendar_att,
            influence_function=calendar_influence_functions,
            keepers=~np.isnan(calendar_att),
            weights=np.ones(n_valid) / n_valid,
            weight_influence_function=None,
        )

        # Overall standard error
        overall_se = _compute_se(
            influence_function=overall_influence_function,
            n_units=n_units,
            estimation_params=estimation_params,
            bootstrap_iterations=bootstrap_iterations,
            alpha=alpha,
        )
    else:
        overall_att = np.nan
        overall_se = np.nan
        overall_influence_function = np.zeros(n_units)

    return AGGTEResult(
        overall_att=overall_att,
        overall_se=overall_se,
        aggregation_type="calendar",
        event_times=original_calendar_times,
        att_by_event=calendar_att,
        se_by_event=calendar_se,
        critical_values=np.full(len(calendar_times), critical_value),
        influence_func=calendar_influence_functions,
        influence_func_overall=overall_influence_function,
        estimation_params=estimation_params,
        call_info={"function": "compute_aggte", "type": "calendar"},
    )


def _compute_weight_inf_func(
    keepers,
    group_probabilities,
    unit_level_weights,
    unit_level_groups,
    groups,
    unique_groups_recoded,
):
    """Compute influence function adjustment for estimated weights.

    This computes the extra term in the influence function that arises
    from having to estimate the weights (group probabilities).

    Parameters
    ----------
    keepers : ndarray
        Indices of observations used in aggregation.
    group_probabilities : ndarray
        Probability of being in each group (length = len(glist)).
    unit_level_weights : ndarray
        Unit-level sampling weights (length = n_units).
    unit_level_groups : ndarray
        Unit-level group assignments (length = n_units).
    groups : ndarray
        Group indicators for ATT(g,t) observations.
    unique_groups_recoded : ndarray
        Unique group values.

    Returns
    -------
    ndarray
        Weight influence function of shape (n_units, len(keepers)).
    """
    if unit_level_groups is None or unit_level_weights is None:
        raise ValueError(
            "Unit-level group assignments (unit_level_groups) and weights (unit_level_weights) are required"
        )

    n_units = len(unit_level_groups)
    n_keepers = len(keepers)

    weight_influence_function = np.zeros((n_units, n_keepers))

    pg_keepers = np.zeros(n_keepers)
    for i, k in enumerate(keepers):
        group_k = groups[k]
        group_idx = np.where(unique_groups_recoded == group_k)[0][0]
        pg_keepers[i] = group_probabilities[group_idx]

    sum_pg = pg_keepers.sum()

    for i, k in enumerate(keepers):
        group_k = groups[k]
        group_idx = np.where(unique_groups_recoded == group_k)[0][0]
        pg_k = group_probabilities[group_idx]

        # Effect of estimating weight in numerator
        # This is unit_level_weights * 1(unit_level_groups_i = g) - group_probabilities[g]
        # for each unit
        if1 = unit_level_weights * (unit_level_groups == group_k).astype(float) - pg_k
        if1 = if1 / sum_pg

        weight_influence_function[:, i] = if1

    # Effect of estimating weights in denominator
    sum_term = np.zeros(n_units)
    for k in keepers:
        group_k = groups[k]
        group_idx = np.where(unique_groups_recoded == group_k)[0][0]
        sum_term += unit_level_weights * (unit_level_groups == group_k).astype(float) - group_probabilities[group_idx]

    weight_influence_function -= np.outer(sum_term / (sum_pg**2), pg_keepers)

    return weight_influence_function


def _get_agg_inf_func(
    att,
    influence_function,
    keepers,
    weights,
    weight_influence_function=None,
):
    """Combine influence functions to get aggregated influence function.

    Parameters
    ----------
    att : ndarray
        Vector of ATT estimates.
    influence_function : ndarray
        Matrix of influence functions (n_units x n_att).
    keepers : ndarray
        Boolean mask or indices of which ATTs to include.
    weights : ndarray
        Weights to apply to each kept ATT.
    weight_influence_function : ndarray, optional
        Weight influence function from estimating group probabilities.

    Returns
    -------
    ndarray
        Aggregated influence function of shape (n_units,).
    """
    if keepers.dtype == bool:
        keep_idx = np.where(keepers)[0]
    else:
        keep_idx = keepers

    influence_function_kept = influence_function[:, keep_idx]
    att_kept = att[keep_idx]

    weighted_inf = influence_function_kept @ weights

    if weight_influence_function is not None:
        weighted_inf += weight_influence_function @ att_kept

    return weighted_inf


def _compute_se(
    influence_function,
    n_units,
    estimation_params,
    bootstrap_iterations,
    alpha,
):
    """Compute standard error from influence function.

    Parameters
    ----------
    influence_function : ndarray
        Influence function vector.
    n_units : int
        Number of cross-sectional units.
    estimation_params : dict
        Dictionary of estimation parameters.
    bootstrap_iterations : int
        Number of bootstrap iterations.
    alpha : float
        Significance level.

    Returns
    -------
    float
        Standard error estimate.
    """
    if estimation_params.get("bootstrap", False):
        # Use multiplier bootstrap
        clustervars = estimation_params.get("clustervars")
        cluster = None
        if clustervars:
            if len(clustervars) > 1:
                raise NotImplementedError("Clustering on multiple variables is not supported yet.")

            cluster_var_name = clustervars[0]

            warnings.warn(
                f"Clustering requested on '{cluster_var_name}' but data not available. "
                "Using standard bootstrap without clustering."
            )

        mboot_result = mboot(
            inf_func=influence_function.reshape(-1, 1),
            n_units=n_units,
            biters=bootstrap_iterations,
            alp=alpha,
            cluster=cluster,
        )
        standard_error = mboot_result["se"][0]
    else:
        # Analytical standard error
        standard_error = np.sqrt(np.mean(influence_function**2) / n_units)

    if standard_error <= np.sqrt(np.finfo(float).eps) * 10:
        standard_error = np.nan

    return standard_error


def _get_z_critical(alpha):
    """Get critical value from standard normal distribution."""
    from scipy import stats

    return stats.norm.ppf(1 - alpha)


def _orig2t(orig, unique_original_times_and_groups, recoded_times):
    """Convert original time value to sequential time value."""
    try:
        idx = np.where(unique_original_times_and_groups == orig)[0][0]
        return recoded_times[idx]
    except IndexError:
        return np.nan


def _t2orig(t, unique_original_times_and_groups, recoded_times):
    """Convert sequential time value back to original time value."""
    try:
        idx = np.where(recoded_times == t)[0][0]
        return unique_original_times_and_groups[idx]
    except IndexError:
        return 0



"""Propensity score estimation using Inverse Probability Tilting (IPT)."""

import warnings

import numpy as np
import scipy.linalg
import scipy.optimize
import scipy.special
import statsmodels.api as sm


def calculate_pscore_ipt(D, X, iw, quantiles=None):
    """Calculate propensity scores using Inverse Probability Tilting.

    Inverse probability tilting (IPT) described in [1]_ involves specifying estimating
    equations that fit the parameters of two or more generalized linear models with a
    modification that ensures exact balance on the co-variate means. These estimating
    equations are solved, and the estimated parameters are used in the (generalized)
    propensity score, which is used to compute the weights.

    Parameters
    ----------
    D : ndarray
        Treatment indicator (1D array).
    X : ndarray
        Covariate matrix (2D array, n_obs x n_features), must include intercept.
    iw : ndarray
        Individual weights (1D array).
    quantiles : dict[int, list[float]] | None
        Dict mapping column indices to quantiles (values between 0 and 1).
        For example, {1: [0.25, 0.5, 0.75]} adds 25th, 50th, 75th percentiles
        of the 2nd column as balance constraints. Default is None (no quantiles).

    Returns
    -------
    ndarray
        Propensity scores.

    References
    ----------
    .. [1] Graham, B., Pinto, C., and Egel, D. (2012), "Inverse Probability Tilting for Moment
        Condition Models with Missing Data," The Review of Economic Studies, 79(3), 1053-1079.
        https://doi.org/10.1093/restud/rdr047
    """
    if np.all(iw == 0):
        warnings.warn(
            "All individual weights are zero. Propensity scores will be uninformative.",
            UserWarning,
        )
        return np.full_like(D, np.nan, dtype=float)

    unique_d = np.unique(D)
    if not np.array_equal(unique_d, [0, 1]):
        warnings.warn(
            f"Treatment indicator D contains values {unique_d}. Expected binary values [0, 1]. "
            "Results may be unreliable.",
            UserWarning,
        )

    X_processed, _ = _remove_collinear_columns(X)

    if quantiles is not None:
        X_processed = _add_quantile_constraints(X_processed, quantiles, iw)

    n_obs, k_features = X_processed.shape
    init_gamma = _get_initial_gamma(D, X_processed, iw, k_features)

    # Try trust-constr optimization first
    try:
        opt_cal_results = scipy.optimize.minimize(
            lambda g, d_arr, x_arr, iw_arr: _loss_ps_cal(g, d_arr, x_arr, iw_arr)[0],
            init_gamma.astype(np.float64),
            args=(D, X_processed, iw),
            method="trust-constr",
            jac=lambda g, d_arr, x_arr, iw_arr: _loss_ps_cal(g, d_arr, x_arr, iw_arr)[1],
            hess=lambda g, d_arr, x_arr, iw_arr: _loss_ps_cal(g, d_arr, x_arr, iw_arr)[2],
            options={"maxiter": 1000},
        )
        if opt_cal_results.success:
            gamma_cal = opt_cal_results.x
        else:
            raise RuntimeError("trust-constr did not converge")
    except (ValueError, np.linalg.LinAlgError, RuntimeError, OverflowError) as e:
        warnings.warn(f"trust-constr optimization failed: {e}. Using IPT algorithm.", UserWarning)

        # Try IPT optimization
        try:
            opt_ipt_results = scipy.optimize.minimize(
                lambda g, d_arr, x_arr, iw_arr, n: _loss_ps_ipt(g, d_arr, x_arr, iw_arr, n)[0],
                init_gamma.astype(np.float64),
                args=(D, X_processed, iw, n_obs),
                method="BFGS",
                jac=lambda g, d_arr, x_arr, iw_arr, n: _loss_ps_ipt(g, d_arr, x_arr, iw_arr, n)[1],
                options={"maxiter": 10000, "gtol": 1e-06},
            )
            if opt_ipt_results.success:
                gamma_cal = opt_ipt_results.x
            else:
                raise RuntimeError("IPT optimization did not converge") from None
        except (ValueError, np.linalg.LinAlgError, RuntimeError, OverflowError) as e_ipt:
            warnings.warn(f"IPT optimization failed: {e_ipt}. Using initial logit estimates.", UserWarning)
            gamma_cal = init_gamma

            # Validate logit fallback
            try:
                logit_model_refit = sm.GLM(D, X_processed, family=sm.families.Binomial(), freq_weights=iw)
                logit_results_refit = logit_model_refit.fit(start_params=init_gamma, maxiter=100)
                if not logit_results_refit.converged:
                    warnings.warn("Initial Logit model (used as fallback) also did not converge.", UserWarning)
            except (ValueError, np.linalg.LinAlgError, RuntimeError, OverflowError):
                warnings.warn("Checking convergence of fallback Logit model failed.", UserWarning)

    # Compute propensity scores
    pscore_linear = X_processed @ gamma_cal
    pscore = scipy.special.expit(pscore_linear)

    if np.any(np.isnan(pscore)):
        warnings.warn(
            "Propensity score model coefficients might have NA/Inf components. "
            "Multicollinearity or lack of variation in covariates is a likely reason. "
            "Resulting pscores contain NaNs.",
            UserWarning,
        )

    return pscore


def _loss_ps_cal(gamma, D, X, iw):
    """Loss function for calibrated propensity score estimation using trust.

    Parameters
    ----------
    gamma : ndarray
        Coefficient vector for propensity score model.
    D : ndarray
        Treatment indicator (1D array).
    X : ndarray
        Covariate matrix (2D array, n_obs x n_features), includes intercept.
    iw : ndarray
        Individual weights (1D array).

    Returns
    -------
    tuple
        (value, gradient, hessian)
    """
    n_obs, k_features = X.shape

    if np.any(np.isnan(gamma)):
        return np.inf, np.full(k_features, np.nan), np.full((k_features, k_features), np.nan)

    ps_ind = X @ gamma
    ps_ind_clipped = np.clip(ps_ind, -500, 500)
    exp_ps_ind = np.exp(ps_ind_clipped)

    value = -np.mean(np.where(D, ps_ind, -exp_ps_ind) * iw)

    grad_terms = np.where(D[:, np.newaxis], 1.0, -exp_ps_ind[:, np.newaxis]) * iw[:, np.newaxis] * X
    gradient = -np.mean(grad_terms, axis=0)

    hess_M_vector = np.where(D, 0.0, -exp_ps_ind) * iw
    hessian_term_matrix = X * hess_M_vector[:, np.newaxis]
    hessian = -(X.T @ hessian_term_matrix) / n_obs
    return value, gradient, hessian


def _loss_ps_ipt(gamma, D, X, iw, n_obs):
    """Loss function for inverse probability tilting propensity score estimation.

    Parameters
    ----------
    gamma : ndarray
        Coefficient vector for propensity score model.
    D : ndarray
        Treatment indicator (1D array).
    X : ndarray
        Covariate matrix (2D array, n_obs x n_features), includes intercept.
    iw : ndarray
        Individual weights (1D array).
    n_obs : int
        Number of observations.

    Returns
    -------
    tuple
        (value, gradient, hessian)
    """
    k_features = X.shape[1]
    if np.any(np.isnan(gamma)):
        return np.inf, np.full(k_features, np.nan), np.full((k_features, k_features), np.nan)

    log_n_minus_1 = np.log(n_obs - 1)
    cn = -(n_obs - 1)
    bn = -n_obs + (n_obs - 1) * log_n_minus_1
    an = -(n_obs - 1) * (1 - log_n_minus_1 + 0.5 * (log_n_minus_1**2))
    v_star = log_n_minus_1

    v = X @ gamma
    v_clipped = np.clip(v, -500, 500)

    phi = np.where(v < v_star, -v - np.exp(v_clipped), an + bn * v + 0.5 * cn * (v**2))
    phi1 = np.where(v < v_star, -1.0 - np.exp(v_clipped), bn + cn * v)
    phi2 = np.where(v < v_star, -np.exp(v_clipped), cn)
    value = -np.sum((iw * (1 - D) * phi) + v)

    grad_vec_term = iw * ((1 - D) * phi1 + 1.0)
    gradient = -(X.T @ grad_vec_term)

    hess_M_ipt_vector = (1 - D) * iw * phi2
    hessian_term_matrix = X * hess_M_ipt_vector[:, np.newaxis]
    hessian = -(hessian_term_matrix.T @ X)
    return value, gradient, hessian


def _get_initial_gamma(D, X, iw, k_features):
    """Get initial gamma values for optimization."""
    try:
        logit_model = sm.GLM(D, X, family=sm.families.Binomial(), freq_weights=iw)
        logit_results = logit_model.fit(maxiter=100)
        init_gamma = logit_results.params
        if not logit_results.converged:
            warnings.warn(
                "Initial Logit model for IPT did not converge. Using pseudo-inverse for initial gamma.", UserWarning
            )
            try:
                init_gamma = np.linalg.pinv(X.T @ (iw[:, np.newaxis] * X)) @ (X.T @ (iw * D))
            except np.linalg.LinAlgError:
                warnings.warn("Pseudo-inverse for initial gamma failed. Using zeros.", UserWarning)
                init_gamma = np.zeros(k_features)
        return init_gamma
    except (np.linalg.LinAlgError, ValueError) as e:
        warnings.warn(f"Initial Logit model failed: {e}. Using zeros for initial gamma.", UserWarning)
        return np.zeros(k_features)


def _remove_collinear_columns(X):
    """Remove collinear columns from covariate matrix."""
    n_obs, n_cols = X.shape

    if n_cols == 1:
        return X, []

    _, R, P = scipy.linalg.qr(X, mode="economic", pivoting=True)

    diag_R = np.abs(np.diag(R))
    tol = diag_R[0] * max(n_obs, n_cols) * np.finfo(float).eps
    rank = np.sum(diag_R > tol)

    if rank == n_cols:
        return X, []

    independent_cols = P[:rank]
    removed_cols = P[rank:]

    if len(removed_cols) > 0:
        warnings.warn(
            f"Removed {len(removed_cols)} collinear column(s) from covariate matrix. "
            f"Column indices removed: {removed_cols.tolist()}",
            UserWarning,
        )

    return X[:, independent_cols], removed_cols.tolist()


def _add_quantile_constraints(X, quantiles, iw):
    """Add quantile balance constraints to covariate matrix."""
    quantile_features = []

    for col_idx, q_list in quantiles.items():
        if col_idx >= X.shape[1]:
            warnings.warn(f"Column index {col_idx} exceeds number of columns ({X.shape[1]}). Skipping.", UserWarning)
            continue

        if col_idx == 0 and np.all(X[:, 0] == X[0, 0]):
            warnings.warn("Skipping quantile constraints for intercept column.", UserWarning)
            continue

        col_values = X[:, col_idx]

        for q in q_list:
            if not 0 < q < 1:
                warnings.warn(f"Quantile {q} must be between 0 and 1. Skipping.", UserWarning)
                continue

            threshold = _weighted_quantile(col_values, q, iw)

            q_indicator = (col_values <= threshold).astype(float)
            quantile_features.append(q_indicator)

    if quantile_features:
        X_extended = np.column_stack([X] + quantile_features)
        return X_extended
    return X


def _weighted_quantile(values, q, weights):
    """Compute weighted quantile."""
    sorted_idx = np.argsort(values)
    sorted_values = values[sorted_idx]
    sorted_weights = weights[sorted_idx]

    # Cumulative weights
    cum_weights = np.cumsum(sorted_weights)
    total_weight = cum_weights[-1]

    # Find quantile position
    quantile_weight = q * total_weight
    idx = np.searchsorted(cum_weights, quantile_weight)

    if idx == 0:
        return sorted_values[0]
    if idx >= len(sorted_values):
        return sorted_values[-1]

    # Linear interpolation between adjacent values
    w_below = cum_weights[idx - 1]
    w_above = cum_weights[idx]

    if w_above == w_below:
        return sorted_values[idx]

    # Interpolation factor
    alpha = (quantile_weight - w_below) / (w_above - w_below)
    return sorted_values[idx - 1] + alpha * (sorted_values[idx] - sorted_values[idx - 1])



"""Preprocessing for multi-period DiD estimators."""

from .builders import DIDDataBuilder
from .constants import (
    BasePeriod,
    ControlGroup,
    DataFormat,
    EstimationMethod,
)
from .models import DIDConfig, DIDData, ValidationResult
from .tensors import TensorFactorySelector
from .transformers import DataTransformerPipeline
from .validators import CompositeValidator

__all__ = [
    # Builders
    "DIDDataBuilder",
    # Constants
    "BasePeriod",
    "ControlGroup",
    "DataFormat",
    "EstimationMethod",
    # Models
    "DIDConfig",
    "DIDData",
    "ValidationResult",
    # Factories and Pipelines
    "TensorFactorySelector",
    "DataTransformerPipeline",
    "CompositeValidator",
]



"""Sensitivity analysis for event study coefficients."""

import warnings
from typing import NamedTuple

import numpy as np
import pandas as pd
from scipy import stats

from .bounds import compute_delta_sd_upperbound_m
from .delta.rm.rm import compute_identified_set_rm
from .fixed_length_ci import compute_flci
from .utils import basis_vector, validate_conformable, validate_symmetric_psd
from .wrappers import DeltaMethodSelector


class SensitivityResult(NamedTuple):
    """Result from sensitivity analysis."""

    lb: float
    ub: float
    method: str
    delta: str
    m: float


class OriginalCSResult(NamedTuple):
    """Result from original confidence set construction."""

    lb: float
    ub: float
    method: str = "Original"
    delta: str | None = None


def create_sensitivity_results_sm(
    betahat,
    sigma,
    num_pre_periods,
    num_post_periods,
    method=None,
    m_vec=None,
    l_vec=None,
    monotonicity_direction=None,
    bias_direction=None,
    alpha=0.05,
    grid_points=1000,
    grid_lb=None,
    grid_ub=None,
):
    r"""Perform sensitivity analysis using smoothness restrictions.

    Implements methods for robust inference in difference-in-differences and event study
    designs using smoothness restrictions on the underlying trend.

    Parameters
    ----------
    betahat : ndarray
        Estimated event study coefficients. Should have length
        num_pre_periods + num_post_periods.
    sigma : ndarray
        Covariance matrix of betahat. Should be
        (num_pre_periods + num_post_periods) x (num_pre_periods + num_post_periods).
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    method : str, optional
        Confidence interval method. Options are:

        - "FLCI": Fixed-length confidence intervals
        - "Conditional": Conditional confidence intervals
        - "C-F": Conditional FLCI hybrid
        - "C-LF": Conditional least-favorable hybrid

        Default is "FLCI" if no restrictions, "C-F" otherwise.
    m_vec : ndarray, optional
        Vector of M values for sensitivity analysis. If None, constructs
        default sequence from 0 to data-driven upper bound.
    l_vec : ndarray, optional
        Vector of weights for parameter of interest. Default is first
        post-period effect.
    monotonicity_direction : str, optional
        Direction of monotonicity restriction: "increasing" or "decreasing".
    bias_direction : str, optional
        Direction of bias restriction: "positive" or "negative".
    alpha : float, default=0.05
        Significance level.
    grid_points : int, default=1000
        Number of grid points for conditional methods.
    grid_lb : float, optional
        Lower bound for grid search. If None, uses data-driven bound.
    grid_ub : float, optional
        Upper bound for grid search. If None, uses data-driven bound.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns: lb, ub, method, Delta, M.

    Notes
    -----
    Cannot specify both monotonicity_direction and bias_direction.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A More Credible Approach to
        Parallel Trends. Review of Economic Studies.
    """
    if l_vec is None:
        l_vec = basis_vector(1, num_post_periods)

    validate_conformable(betahat, sigma, num_pre_periods, num_post_periods, l_vec)
    validate_symmetric_psd(sigma)

    if monotonicity_direction is not None and bias_direction is not None:
        raise ValueError("Cannot specify both monotonicity_direction and bias_direction.")

    if m_vec is None:
        if num_pre_periods == 1:
            # With only one pre-period, we can't estimate second differences
            # so use a simple range based on the pre-period variance
            m_vec = np.linspace(0, np.sqrt(sigma[0, 0]), 10)
        else:
            # Use a data-driven upper bound based on pre-treatment variation
            m_ub = compute_delta_sd_upperbound_m(
                betahat=betahat,
                sigma=sigma,
                num_pre_periods=num_pre_periods,
                alpha=0.05,
            )
            m_vec = np.linspace(0, m_ub, 10)

    results = []

    compute_fn, delta_type = DeltaMethodSelector.get_smoothness_method(
        monotonicity_direction=monotonicity_direction,
        bias_direction=bias_direction,
    )

    if method is None:
        if monotonicity_direction is None and bias_direction is None:
            method = "FLCI"
        else:
            method = "C-F"

    if method == "FLCI" and (monotonicity_direction is not None or bias_direction is not None):
        warnings.warn(
            "You specified a shape/sign restriction but method = FLCI. The FLCI does not use these restrictions!"
        )

    for m in m_vec:
        if method == "FLCI":
            # Fixed-length CI doesn't incorporate shape restrictions
            flci_result = compute_flci(
                beta_hat=betahat,
                sigma=sigma,
                n_pre_periods=num_pre_periods,
                n_post_periods=num_post_periods,
                post_period_weights=l_vec,
                smoothness_bound=m,
                alpha=alpha,
            )
            results.append(
                SensitivityResult(
                    lb=flci_result.flci[0],
                    ub=flci_result.flci[1],
                    method="FLCI",
                    delta=delta_type,
                    m=m,
                )
            )
        elif method in ["Conditional", "C-F", "C-LF"]:
            hybrid_flag = {
                "Conditional": "ARP",  # Andrews, Roth, Pakes (2022)
                "C-F": "FLCI",  # Conditional + FLCI hybrid
                "C-LF": "LF",  # Conditional + Least Favorable hybrid
            }[method]

            delta_kwargs = {
                "betahat": betahat,
                "sigma": sigma,
                "num_pre_periods": num_pre_periods,
                "num_post_periods": num_post_periods,
                "l_vec": l_vec,
                "alpha": alpha,
                "m_bar": m,
                "hybrid_flag": hybrid_flag,
                "grid_points": grid_points,
                "grid_lb": grid_lb,
                "grid_ub": grid_ub,
            }

            if monotonicity_direction is not None:
                delta_kwargs["monotonicity_direction"] = monotonicity_direction
            elif bias_direction is not None:
                delta_kwargs["bias_direction"] = bias_direction

            cs_result = compute_fn(**delta_kwargs)
            accept_idx = np.where(cs_result["accept"])[0]
            if len(accept_idx) > 0:
                lb = cs_result["grid"][accept_idx[0]]
                ub = cs_result["grid"][accept_idx[-1]]
            else:
                lb = np.nan
                ub = np.nan

            results.append(SensitivityResult(lb=lb, ub=ub, method=method, delta=delta_type, m=m))
        else:
            raise ValueError(f"Unknown method: {method}")

    df = pd.DataFrame(results)
    return df


def create_sensitivity_results_rm(
    betahat,
    sigma,
    num_pre_periods,
    num_post_periods,
    bound="deviation from parallel trends",
    method="C-LF",
    m_bar_vec=None,
    l_vec=None,
    monotonicity_direction=None,
    bias_direction=None,
    alpha=0.05,
    grid_points=1000,
    grid_lb=None,
    grid_ub=None,
):
    r"""Perform sensitivity analysis using relative magnitude bounds.

    Implements methods for robust inference using bounds on the relative magnitude
    of post-treatment violations compared to pre-treatment violations.

    Parameters
    ----------
    betahat : ndarray
        Estimated event study coefficients.
    sigma : ndarray
        Covariance matrix of betahat.
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    bound : str, default="deviation from parallel trends"
        Type of bound:

        - "Deviation from parallel trends": :math:`\Delta^{RM}` and variants
        - "Deviation from linear trend": :math:`\Delta^{SDRM}` and variants

    method : str, default="C-LF"
        Confidence interval method: "Conditional" or "C-LF".
    m_bar_vec : ndarray, optional
        Vector of :math:`\bar{M}` values. Default is 10 values from 0 to 2.
    l_vec : ndarray, optional
        Vector of weights for parameter of interest.
    monotonicity_direction : str, optional
        Direction of monotonicity restriction: "increasing" or "decreasing".
    bias_direction : str, optional
        Direction of bias restriction: "positive" or "negative".
    alpha : float, default=0.05
        Significance level.
    grid_points : int, default=1000
        Number of grid points for conditional methods.
    grid_lb : float, optional
        Lower bound for grid search.
    grid_ub : float, optional
        Upper bound for grid search.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns: lb, ub, method, Delta, Mbar.

    Notes
    -----
    Deviation from linear trend requires at least 3 pre-treatment periods.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A More Credible Approach to
        Parallel Trends. Review of Economic Studies.
    """
    if l_vec is None:
        l_vec = basis_vector(1, num_post_periods)

    validate_conformable(betahat, sigma, num_pre_periods, num_post_periods, l_vec)
    validate_symmetric_psd(sigma)

    if bound not in ["deviation from parallel trends", "deviation from linear trend"]:
        raise ValueError("bound must be 'deviation from parallel trends' or 'deviation from linear trend'")

    if monotonicity_direction is not None and bias_direction is not None:
        raise ValueError("Cannot specify both monotonicity_direction and bias_direction.")

    if method not in ["Conditional", "C-LF"]:
        raise ValueError("method must be 'Conditional' or 'C-LF'")

    if m_bar_vec is None:
        # Default grid for relative magnitude parameter
        # 0 = parallel trends, 1 = same magnitude violations allowed, 2 = twice as large
        m_bar_vec = np.linspace(0, 2, 10)

    hybrid_flag = "ARP" if method == "Conditional" else "LF"

    results = []

    if bound == "deviation from linear trend" and num_pre_periods < 3:
        raise ValueError(
            "Not enough pre-periods for 'deviation from linear trend' (Delta^SDRM requires at least 3 pre-periods)"
        )

    compute_fn, delta_type = DeltaMethodSelector.get_relative_magnitude_method(
        bound_type=bound,
        monotonicity_direction=monotonicity_direction,
        bias_direction=bias_direction,
    )

    for m_bar in m_bar_vec:
        # If grid bounds are not user-specified, calculate them based on the identified set for this Mbar
        if grid_lb is None or grid_ub is None:
            id_set = compute_identified_set_rm(
                m_bar=m_bar,
                true_beta=betahat,
                l_vec=l_vec,
                num_pre_periods=num_pre_periods,
                num_post_periods=num_post_periods,
            )
            post_indices = slice(num_pre_periods, num_pre_periods + num_post_periods)
            sd_theta = np.sqrt(l_vec.flatten() @ sigma[post_indices, post_indices] @ l_vec.flatten())

            current_grid_lb = id_set.id_lb - 20 * sd_theta
            current_grid_ub = id_set.id_ub + 20 * sd_theta
        else:
            current_grid_lb = grid_lb
            current_grid_ub = grid_ub

        delta_kwargs = {
            "betahat": betahat,
            "sigma": sigma,
            "num_pre_periods": num_pre_periods,
            "num_post_periods": num_post_periods,
            "l_vec": l_vec,
            "alpha": alpha,
            "m_bar": m_bar,
            "hybrid_flag": hybrid_flag,
            "grid_points": grid_points,
            "grid_lb": current_grid_lb,
            "grid_ub": current_grid_ub,
        }

        if monotonicity_direction is not None:
            delta_kwargs["monotonicity_direction"] = monotonicity_direction
        elif bias_direction is not None:
            delta_kwargs["bias_direction"] = bias_direction

        cs_result = compute_fn(**delta_kwargs)
        accept_idx = np.where(cs_result["accept"])[0]
        if len(accept_idx) > 0:
            lb = cs_result["grid"][accept_idx[0]]
            ub = cs_result["grid"][accept_idx[-1]]
        else:
            lb = np.nan
            ub = np.nan

        results.append(SensitivityResult(lb=lb, ub=ub, method=method, delta=delta_type, m=m_bar))

    df = pd.DataFrame(results)
    df = df.rename(columns={"m": "Mbar"})
    return df


def construct_original_cs(
    betahat,
    sigma,
    num_pre_periods,
    num_post_periods,
    l_vec=None,
    alpha=0.05,
):
    r"""Construct original (non-robust) confidence set.

    Constructs a standard confidence interval for the parameter of interest
    without any robustness to violations of parallel trends.

    Parameters
    ----------
    betahat : ndarray
        Estimated event study coefficients.
    sigma : ndarray
        Covariance matrix of betahat.
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    l_vec : ndarray, optional
        Vector of weights for parameter of interest. Default is first
        post-period effect.
    alpha : float, default=0.05
        Significance level.

    Returns
    -------
    OriginalCSResult
        NamedTuple with lb, ub, method="Original", delta=None.
    """
    if l_vec is None:
        l_vec = basis_vector(1, num_post_periods)

    validate_conformable(betahat, sigma, num_pre_periods, num_post_periods, l_vec)
    validate_symmetric_psd(sigma)

    post_beta = betahat[num_pre_periods:]
    post_sigma = sigma[num_pre_periods:, num_pre_periods:]

    l_vec_flat = l_vec.flatten()
    se = np.sqrt(l_vec_flat @ post_sigma @ l_vec_flat)

    point_est = l_vec_flat @ post_beta

    z_alpha = stats.norm.ppf(1 - alpha / 2)
    lb = point_est - z_alpha * se
    ub = point_est + z_alpha * se

    return OriginalCSResult(lb=lb, ub=ub)



"""Utility functions for drdid estimators."""

import warnings

import formulaic as fml
import numpy as np
import pandas as pd


def preprocess_drdid(
    data,
    y_col,
    time_col,
    id_col,
    treat_col,
    covariates_formula=None,
    panel=True,
    normalized=True,
    est_method="imp",
    weights_col=None,
    boot=False,
    boot_type="weighted",
    n_boot=None,
    inf_func=False,
):
    """Pre-processes data for DR DiD estimation.

    Validates input data, checks for required columns, handles missing values,
    balances panel data if requested, checks for time-invariant treatment/covariates
    in panel data, processes covariates using patsy, normalizes weights, and
    structures the output for downstream estimation functions.

    Parameters
    ----------
    data : pd.DataFrame
        The input data containing outcome, time, unit ID, treatment,
        and optionally covariates and weights. Must contain exactly two time periods.
    y_col : str
        Name of the column in `data` representing the outcome variable.
    time_col : str
        Name of the column in `data` representing the time period indicator.
        Must contain exactly two unique numeric values.
    id_col : str
        Name of the column in `data` representing the unique unit identifier.
        Required if `panel=True`.
    treat_col : str
        Name of the column in `data` representing the treatment indicator.
        Must contain only 0 (control) and 1 (treated). For panel data, this
        should indicate whether a unit is *ever* treated (time-invariant).
        For repeated cross-sections, it indicates treatment status in the post-period.
    covariates_formula : str or None, default None
        A patsy-style formula string for specifying covariates (e.g., "~ x1 + x2 + x1:x2").
        If None, only an intercept term (`~ 1`) is included. Covariates specified
        here must exist as columns in `data`. For panel data, covariates must be
        time-invariant.
    panel : bool, default True
        Indicates whether the data represents panel observations (True) or
        repeated cross-sections (False). If True, data is balanced, and
        treatment/covariates/weights are checked for time-invariance.
    normalized : bool, default True
        If True, the observation weights (`weights_col` or unit weights if None)
        are normalized to have a mean of 1.
    est_method : {"imp", "trad"}, default "imp"
        Specifies the estimation method context, potentially influencing future
        preprocessing steps (currently informational). "imp" for imputation-based,
        "trad" for traditional regression-based.
    weights_col : str or None, default None
        Name of the column in `data` containing observation weights. If None,
        unit weights (all 1.0) are assumed. For panel data, weights must be
        time-invariant.
    boot : bool, default False
        Flag indicating whether preprocessing is done in preparation for a
        bootstrap procedure (currently informational).
    boot_type : {"weighted", "multiplier"}, default "weighted"
        Specifies the type of bootstrap procedure if `boot=True` (currently informational).
    n_boot : int or None, default None
        Number of bootstrap replications if `boot=True` (currently informational).
    inf_func : bool, default False
        Flag indicating whether preprocessing is done for influence function
        calculations (currently informational).

    Returns
    -------
    dict[str, Any]
        A dictionary containing processed data elements.
    """
    if not isinstance(data, pd.DataFrame):
        warnings.warn("Input data is not a pandas DataFrame; converting...", UserWarning)
        df = pd.DataFrame(data)
    else:
        df = data.copy()

    required_cols = [y_col, time_col, treat_col]
    if panel:
        required_cols.append(id_col)
    if weights_col:
        required_cols.append(weights_col)

    missing_cols = [col for col in required_cols if col not in df.columns]
    if missing_cols:
        raise ValueError(f"Missing required columns: {', '.join(missing_cols)}")

    if est_method not in ["imp", "trad"]:
        warnings.warn(f"est_method='{est_method}' is not supported. Using 'imp'.", UserWarning)
        est_method = "imp"

    if boot and boot_type not in ["weighted", "multiplier"]:
        warnings.warn(f"boot_type='{boot_type}' is not supported. Using 'weighted'.", UserWarning)
        boot_type = "weighted"

    if not isinstance(normalized, bool):
        warnings.warn(f"normalized={normalized} is not supported. Using True.", UserWarning)
        normalized = True

    numeric_cols = [y_col, time_col, treat_col]
    if panel:
        numeric_cols.append(id_col)
    if weights_col:
        numeric_cols.append(weights_col)

    for col in numeric_cols:
        if not pd.api.types.is_numeric_dtype(df[col]):
            try:
                df[col] = pd.to_numeric(df[col])
                warnings.warn(f"Column '{col}' was not numeric; converted successfully.", UserWarning)
            except (ValueError, TypeError) as exc:
                raise TypeError(f"Column '{col}' must be numeric. Could not convert.") from exc

    time_periods = sorted(df[time_col].unique())
    if len(time_periods) != 2:
        raise ValueError("This package currently supports only two time periods (pre and post).")
    pre_period, post_period = time_periods

    groups = sorted(df[treat_col].unique())
    if len(groups) != 2 or not all(g in [0, 1] for g in groups):
        raise ValueError("Treatment indicator column must contain only 0 (control) and 1 (treated).")

    if covariates_formula is None:
        covariates_formula = "~ 1"

    try:
        model_matrix_result = fml.model_matrix(
            covariates_formula,
            df,
            output="pandas",
        )
        covariates_df = model_matrix_result
        if hasattr(model_matrix_result, "model_spec") and model_matrix_result.model_spec:
            original_cov_names = [var for var in model_matrix_result.model_spec.variables if var != "1"]
        else:
            original_cov_names = []
            warnings.warn("Could not retrieve model_spec from formulaic output. ", UserWarning)

    except Exception as e:
        raise ValueError(f"Error processing covariates_formula '{covariates_formula}' with formulaic: {e}") from e

    cols_to_drop = [name for name in original_cov_names if name in df.columns]
    df_processed = pd.concat([df.drop(columns=cols_to_drop), covariates_df], axis=1)

    if weights_col:
        df_processed["weights"] = df_processed[weights_col]
        if not pd.api.types.is_numeric_dtype(df_processed["weights"]):
            raise TypeError(f"Weights column '{weights_col}' must be numeric.")
        if (df_processed["weights"] < 0).any():
            warnings.warn("Some weights are negative. Ensure this is intended.", UserWarning)
    else:
        df_processed["weights"] = 1.0

    initial_rows = len(df_processed)
    cols_for_na_check_base = [y_col, time_col, "weights"] + list(covariates_df.columns)
    if pd.api.types.is_numeric_dtype(df_processed[treat_col]):
        cols_for_na_check_base.append(treat_col)
    if panel:
        cols_for_na_check = cols_for_na_check_base + [id_col]
    else:
        cols_for_na_check = cols_for_na_check_base

    na_counts = df_processed[cols_for_na_check].isna().sum()
    cols_with_na = na_counts[na_counts > 0].index.to_list()

    if cols_with_na:
        warnings.warn(
            f"Missing values found in columns: {', '.join(cols_with_na)}. "
            "Dropping rows with any missing values in relevant columns.",
            UserWarning,
        )
        df_processed = df_processed.dropna(subset=cols_for_na_check)

    if len(df_processed) < initial_rows:
        warnings.warn(f"Dropped {initial_rows - len(df_processed)} rows due to missing values.", UserWarning)
    if df_processed.empty:
        raise ValueError("DataFrame is empty after handling missing values.")

    unique_treat_values = df_processed[treat_col].unique()
    if len(unique_treat_values) < 2:
        raise ValueError(
            f"Data must contain both treated (1) and control (0) units in '{treat_col}'. "
            f"Found only: {unique_treat_values}. "
            "Ensure both groups are present after NA handling."
        )
    if not (np.any(unique_treat_values == 0) and np.any(unique_treat_values == 1)):
        raise ValueError(
            f"Treatment indicator column '{treat_col}' must contain both 0 and 1. "
            f"Found values: {unique_treat_values}. "
            "Ensure both groups are present after NA handling."
        )

    if panel:
        if df_processed.groupby([id_col, time_col]).size().max() > 1:
            raise ValueError(f"ID '{id_col}' is not unique within time period '{time_col}'.")

        _check_treatment_uniqueness(df_processed, id_col, treat_col)

        df_processed = _make_balanced_panel(df_processed, id_col, time_col)
        if df_processed.empty:
            raise ValueError("Balancing the panel resulted in an empty DataFrame. Check input data.")

        df_processed = df_processed.sort_values(by=[id_col, time_col])

        pre_df = df_processed[df_processed[time_col] == pre_period].set_index(id_col)
        post_df = df_processed[df_processed[time_col] == post_period].set_index(id_col)

        common_ids = pre_df.index.intersection(post_df.index)
        pre_df = pre_df.loc[common_ids]
        post_df = post_df.loc[common_ids]

        cov_cols_to_check = [col for col in covariates_df.columns if col != "Intercept"]
        if cov_cols_to_check:
            if not pre_df[cov_cols_to_check].equals(post_df[cov_cols_to_check]):
                diff_mask = (pre_df[cov_cols_to_check] != post_df[cov_cols_to_check]).any()
                diff_cols = diff_mask[diff_mask].index.to_list()
                raise ValueError(f"Covariates must be time-invariant in panel data. Differing columns: {diff_cols}")

        if not pre_df[treat_col].equals(post_df[treat_col]):
            raise ValueError(f"Treatment indicator ('{treat_col}') must be time-invariant in panel data.")

        if not pre_df["weights"].equals(post_df["weights"]):
            raise ValueError("Weights must be time-invariant in panel data.")

    covariates_final = df_processed[covariates_df.columns].values
    if covariates_final.shape[1] > 1:
        _, r = np.linalg.qr(covariates_final)
        diag_r = np.abs(np.diag(r))
        tol = diag_r.max() * 1e-6
        rank = np.sum(diag_r > tol)
        num_covariates = covariates_final.shape[1]

        if rank < num_covariates:
            warnings.warn(
                "Potential collinearity detected among covariates. "
                f"Rank ({rank}) is less than number of covariates ({num_covariates}). "
                "Results may be unstable.",
                UserWarning,
            )

    min_obs_per_group_period = df_processed.groupby([treat_col, time_col]).size().min()
    req_size = covariates_final.shape[1] + 5
    if min_obs_per_group_period < req_size:
        warnings.warn(
            "Small group size detected. Minimum observations in a treatment/period group is "
            f"{min_obs_per_group_period}, which might be less than recommended ({req_size}). "
            "Inference may be unreliable.",
            UserWarning,
        )

    rename_dict = {y_col: "y", treat_col: "D", time_col: "time"}
    if panel:
        rename_dict[id_col] = "id"
    df_processed = df_processed.rename(columns=rename_dict)

    if normalized and "weights" in df_processed.columns:
        mean_weight = df_processed["weights"].mean()
        if mean_weight > 0:
            df_processed["weights"] = df_processed["weights"] / mean_weight
        else:
            warnings.warn("Mean of weights is zero or negative. Cannot normalize.", UserWarning)

    output = {
        "panel": panel,
        "est_method": est_method,
        "normalized": normalized,
        "boot": boot,
        "boot_type": boot_type,
        "n_boot": n_boot,
        "inf_func": inf_func,
        "covariate_names": list(covariates_df.columns),
    }

    if panel:
        df_processed = df_processed.sort_values(by=["id", "time"])
        post_data = df_processed[df_processed["time"] == post_period].set_index("id")
        pre_data = df_processed[df_processed["time"] == pre_period].set_index("id")
        common_ids = post_data.index.intersection(pre_data.index)
        post_data = post_data.loc[common_ids]
        pre_data = pre_data.loc[common_ids]

        output["y1"] = post_data["y"].values
        output["y0"] = pre_data["y"].values
        output["D"] = post_data["D"].values
        output["covariates"] = post_data[output["covariate_names"]].values
        output["weights"] = post_data["weights"].values
        output["n_units"] = len(common_ids)

    else:
        output["y"] = df_processed["y"].values
        output["D"] = df_processed["D"].values
        output["post"] = (df_processed["time"] == post_period).astype(int).values
        output["covariates"] = df_processed[output["covariate_names"]].values
        output["weights"] = df_processed["weights"].values
        output["n_obs"] = len(df_processed)
    return output


def _check_treatment_uniqueness(df, id_col, treat_col):
    """Check if treatment status is unique for each ID in panel data."""
    treat_counts = df.groupby(id_col)[treat_col].nunique()
    if (treat_counts > 1).any():
        invalid_ids = treat_counts[treat_counts > 1].index.to_list()
        raise ValueError(
            f"Treatment indicator ('{treat_col}') must be unique for each ID ('{id_col}'). "
            f"IDs with varying treatment: {invalid_ids}."
        )


def _make_balanced_panel(df, id_col, time_col):
    """Convert an unbalanced panel DataFrame into a balanced one."""
    n_times = df[time_col].nunique()
    obs_counts = df.groupby(id_col).size()
    ids_to_keep = obs_counts[obs_counts == n_times].index
    if len(ids_to_keep) < len(obs_counts):
        warnings.warn(
            "Panel data is unbalanced. Dropping units with incomplete observations.",
            UserWarning,
        )
    return df[df[id_col].isin(ids_to_keep)].copy()


def _validate_inputs(arrays_dict, x, n_bootstrap, trim_level, check_intercept=False):
    """Validate inputs for bootstrap functions."""
    for name, arr in arrays_dict.items():
        if not isinstance(arr, np.ndarray):
            raise TypeError(f"{name} must be a NumPy array.")

    if not isinstance(x, np.ndarray):
        raise TypeError("x must be a NumPy array.")

    for name, arr in arrays_dict.items():
        if arr.ndim != 1:
            raise ValueError(f"{name} must be 1-dimensional.")

    if x.ndim != 2:
        raise ValueError("x must be a 2-dimensional array.")

    first_array = next(iter(arrays_dict.values()))
    n_units = first_array.shape[0]

    for name, arr in arrays_dict.items():
        if arr.shape[0] != n_units:
            raise ValueError("All arrays must have the same number of observations.")

    if x.shape[0] != n_units:
        raise ValueError("All arrays must have the same number of observations.")

    if not isinstance(n_bootstrap, int) or n_bootstrap <= 0:
        raise ValueError("n_bootstrap must be a positive integer.")

    if not 0 < trim_level < 1:
        raise ValueError("trim_level must be between 0 and 1.")

    if check_intercept and not np.all(x[:, 0] == 1.0):
        warnings.warn(
            "The first column of the covariate matrix 'x' does not appear to be an intercept (all ones). "
            "IPT propensity score estimation typically requires an intercept.",
            UserWarning,
        )

    return n_units


def _validate_wols_arrays(arrays_dict, x, function_name="wols"):
    """Validate input arrays for WOLS functions."""
    all_arrays = list(arrays_dict.values()) + [x]
    if not all(isinstance(arr, np.ndarray) for arr in all_arrays):
        raise TypeError("All inputs must be NumPy arrays.")

    if function_name == "wols_panel":
        dim_error_msg = "delta_y, d, ps, and i_weights must be 1-dimensional."
    else:  # wols_rc
        dim_error_msg = "y, post, d, ps, and i_weights must be 1-dimensional."

    for arr in arrays_dict.values():
        if arr.ndim != 1:
            raise ValueError(dim_error_msg)

    if x.ndim != 2:
        raise ValueError("x must be a 2-dimensional array.")

    n_units = next(iter(arrays_dict.values())).shape[0]
    for arr in list(arrays_dict.values()) + [x]:
        if arr.shape[0] != n_units:
            raise ValueError("All arrays must have the same number of observations (first dimension).")

    return n_units


def _check_extreme_weights(weights, threshold=1e6):
    """Check for extreme weight ratios and warn if found."""
    if len(weights) > 1:
        positive_mask = weights > 0
        if np.any(positive_mask):
            min_positive = np.min(weights[positive_mask])
            max_weight = np.max(weights)
            if max_weight / min_positive > threshold:
                warnings.warn("Extreme weight ratios detected. Results may be numerically unstable.", UserWarning)


def _check_wls_condition_number(results, threshold_error=1e15, threshold_warn=1e10):
    """Check condition number of WLS results and handle accordingly."""
    try:
        condition_number = results.condition_number
        if condition_number > threshold_error:
            raise ValueError(
                f"Failed to solve linear system: The covariate matrix may be singular or ill-conditioned "
                f"(condition number: {condition_number:.2e})."
            )
        if condition_number > threshold_warn:
            warnings.warn(
                f"Potential multicollinearity detected (condition number: {condition_number:.2e}). "
                "Consider removing or combining covariates.",
                UserWarning,
            )
    except AttributeError:
        pass


def _check_coefficients_validity(coefficients):
    """Check if coefficients contain invalid values."""
    if np.any(np.isnan(coefficients)) or np.any(np.isinf(coefficients)):
        raise ValueError(
            "Failed to solve linear system. Coefficients contain NaN/Inf values, "
            "likely due to multicollinearity or singular matrix."
        )


def _weighted_sum(term_val, weight_val, term_name):
    sum_w = np.sum(weight_val)
    if sum_w == 0 or not np.isfinite(sum_w):
        warnings.warn(f"Sum of weights for {term_name} is {sum_w}. Term will be NaN.", UserWarning)
        return np.nan

    weighted_sum_term = np.sum(weight_val * term_val)
    if not np.isfinite(weighted_sum_term):
        warnings.warn(
            f"Weighted sum for {term_name} is not finite ({weighted_sum_term}). Term will be NaN.", UserWarning
        )
        return np.nan

    return weighted_sum_term / sum_w



"""Second differences delta methods."""

from .sd import compute_conditional_cs_sd
from .sdb import compute_conditional_cs_sdb
from .sdm import compute_conditional_cs_sdm

__all__ = [
    "compute_conditional_cs_sd",
    "compute_conditional_cs_sdb",
    "compute_conditional_cs_sdm",
]



"""Aggregate Treatment Effect Parameters Object."""

from typing import Literal, NamedTuple

import numpy as np


class AGGTEResult(NamedTuple):
    """Container for aggregated treatment effect parameters.

    Attributes
    ----------
    overall_att : float
        The estimated overall average treatment effect on the treated.
    overall_se : float
        Standard error for overall ATT.
    aggregation_type : {'simple', 'dynamic', 'group', 'calendar'}
        Type of aggregation performed.
    event_times : np.ndarray, optional
        Event/group/time values depending on aggregation type:

        - For dynamic effects: length of exposure
        - For group effects: treatment group indicators
        - For calendar effects: time periods
    att_by_event : ndarray, optional
        ATT estimates specific to each event time value.
    se_by_event : ndarray, optional
        Standard errors specific to each event time value.
    critical_values : ndarray, optional
        Critical values for uniform confidence bands.
    influence_func : ndarray, optional
        Influence function of the aggregated parameters.

        - For overall ATT: 1D array of length n_units
        - For dynamic/group/calendar: 2D array of shape (n_units, n_events) containing
          influence functions for each event-specific ATT
    influence_func_overall : ndarray, optional
        Influence function for the overall ATT (1D array of length n_units).
        This is stored separately for compatibility with both aggregation types.
    min_event_time : int, optional
        Minimum event time (for dynamic effects).
    max_event_time : int, optional
        Maximum event time (for dynamic effects).
    balanced_event_threshold : int, optional
        Balanced event time threshold.
    estimation_params : dict
        Dictionary containing DID estimation parameters including:

        - alpha: significance level
        - bootstrap: whether bootstrap was used
        - uniform_bands: whether uniform confidence bands were computed
        - control_group: 'nevertreated' or 'notyettreated'
        - anticipation_periods: number of anticipation periods
        - estimation_method: estimation method used
    call_info : dict
        Information about the function call that created this object.
    """

    overall_att: float
    overall_se: float
    aggregation_type: Literal["simple", "dynamic", "group", "calendar"]
    event_times: np.ndarray | None = None
    att_by_event: np.ndarray | None = None
    se_by_event: np.ndarray | None = None
    critical_values: np.ndarray | None = None
    influence_func: np.ndarray | None = None
    influence_func_overall: np.ndarray | None = None
    min_event_time: int | None = None
    max_event_time: int | None = None
    balanced_event_threshold: int | None = None
    estimation_params: dict = {}
    call_info: dict = {}


def aggte(
    overall_att,
    overall_se,
    aggregation_type="simple",
    event_times=None,
    att_by_event=None,
    se_by_event=None,
    critical_values=None,
    influence_func=None,
    influence_func_overall=None,
    min_event_time=None,
    max_event_time=None,
    balanced_event_threshold=None,
    estimation_params=None,
    call_info=None,
):
    """Create an aggregate treatment effect result object.

    Parameters
    ----------
    overall_att : float
        The estimated overall ATT.
    overall_se : float
        Standard error for overall ATT.
    aggregation_type : {'simple', 'dynamic', 'group', 'calendar'}, default='simple'
        Type of aggregation performed.
    event_times : ndarray, optional
        Event/group/time values for disaggregated effects.
    att_by_event : ndarray, optional
        ATT estimates for each event time value.
    se_by_event : ndarray, optional
        Standard errors for each event time value.
    critical_values : ndarray, optional
        Critical values for confidence bands.
    influence_func : ndarray, optional
        Influence function of aggregated parameters.

        - For dynamic/group/calendar: 2D array of shape (n_units, n_events)
        - For simple: 1D array of length n_units
    influence_func_overall : ndarray, optional
        Influence function for the overall ATT (1D array).
    min_event_time : int, optional
        Minimum event time.
    max_event_time : int, optional
        Maximum event time.
    balanced_event_threshold : int, optional
        Balanced event time threshold.
    estimation_params : dict, optional
        DID estimation parameters.
    call_info : dict, optional
        Information about the function call.

    Returns
    -------
    AGGTEResult
        NamedTuple containing aggregated treatment effect parameters.
    """
    if aggregation_type not in ["simple", "dynamic", "group", "calendar"]:
        raise ValueError(
            f"Invalid aggregation_type: {aggregation_type}. Must be one of 'simple', 'dynamic', 'group', 'calendar'."
        )

    if event_times is not None:
        n_events = len(event_times)
        if att_by_event is not None and len(att_by_event) != n_events:
            raise ValueError("att_by_event must have same length as event_times.")
        if se_by_event is not None and len(se_by_event) != n_events:
            raise ValueError("se_by_event must have same length as event_times.")
        if critical_values is not None and len(critical_values) != n_events:
            raise ValueError("critical_values must have same length as event_times.")

    if estimation_params is None:
        estimation_params = {}
    if call_info is None:
        call_info = {}

    return AGGTEResult(
        overall_att=overall_att,
        overall_se=overall_se,
        aggregation_type=aggregation_type,
        event_times=event_times,
        att_by_event=att_by_event,
        se_by_event=se_by_event,
        critical_values=critical_values,
        influence_func=influence_func,
        influence_func_overall=influence_func_overall,
        min_event_time=min_event_time,
        max_event_time=max_event_time,
        balanced_event_threshold=balanced_event_threshold,
        estimation_params=estimation_params,
        call_info=call_info,
    )


def format_aggte_result(result):
    """Format aggregate treatment effect results.

    Parameters
    ----------
    result : AGGTEResult
        The aggregate treatment effect result to format.

    Returns
    -------
    str
        Formatted string representation of the results.
    """
    lines = []

    lines.append("=" * 78)
    if result.aggregation_type == "dynamic":
        lines.append(" Aggregate Treatment Effects (Event Study)")
    elif result.aggregation_type == "group":
        lines.append(" Aggregate Treatment Effects (Group/Cohort)")
    elif result.aggregation_type == "calendar":
        lines.append(" Aggregate Treatment Effects (Calendar Time)")
    else:
        lines.append(" Aggregate Treatment Effects")
    lines.append("=" * 78)

    if result.call_info:
        lines.append("")
        lines.append(" Call:")
        if "function" in result.call_info:
            lines.append(f"   {result.call_info['function']}")

    lines.append("")
    if result.aggregation_type == "dynamic":
        lines.append(" Overall summary of ATT's based on event-study/dynamic aggregation:")
    elif result.aggregation_type == "group":
        lines.append(" Overall summary of ATT's based on group/cohort aggregation:")
    elif result.aggregation_type == "calendar":
        lines.append(" Overall summary of ATT's based on calendar time aggregation:")
    else:
        lines.append(" Overall ATT:")

    alpha = result.estimation_params.get("alpha", 0.05)
    conf_level = int((1 - alpha) * 100)

    z_crit = _get_z_critical(alpha / 2)
    overall_lci = result.overall_att - z_crit * result.overall_se
    overall_uci = result.overall_att + z_crit * result.overall_se

    overall_sig = (overall_uci < 0) or (overall_lci > 0)
    sig_marker = "*" if overall_sig else " "

    lines.append("")
    lines.append(f"   ATT          Std. Error     [{conf_level}% Conf. Interval]")
    lines.append(
        f"   {result.overall_att:10.4f}   {result.overall_se:10.4f}     [{overall_lci:7.4f}, {overall_uci:7.4f}] {sig_marker}"  # noqa: E501
    )

    if result.aggregation_type in ["dynamic", "group", "calendar"] and result.event_times is not None:
        lines.append("")
        lines.append("")

        if result.aggregation_type == "dynamic":
            lines.append(" Dynamic Effects:")
            col1_header = "Event time"
        elif result.aggregation_type == "group":
            lines.append(" Group Effects:")
            col1_header = "Group"
        else:  # calendar
            lines.append(" Time Effects:")
            col1_header = "Time"

        lines.append("")

        bootstrap = result.estimation_params.get("bootstrap", False)
        uniform_bands = result.estimation_params.get("uniform_bands", False)
        if bootstrap and uniform_bands:
            cb_type = f"[{conf_level}% Simult. Conf. Band]"
        else:
            cb_type = f"[{conf_level}% Pointwise Conf. Band]"

        lines.append(f"  {col1_header:>12}   Estimate   Std. Error   {cb_type}")

        if result.critical_values is not None:
            crit_vals = result.critical_values
        else:
            crit_vals = np.full(len(result.event_times), z_crit)

        lower_bounds = result.att_by_event - crit_vals * result.se_by_event
        upper_bounds = result.att_by_event + crit_vals * result.se_by_event

        for event_val, att_val, se_val, lb, ub in zip(
            result.event_times, result.att_by_event, result.se_by_event, lower_bounds, upper_bounds
        ):
            sig = (ub < 0) or (lb > 0)
            sig_marker = "*" if sig else " "

            lines.append(
                f"  {event_val:>12.0f}   {att_val:8.4f}   {se_val:10.4f}   [{lb:7.4f}, {ub:7.4f}] {sig_marker}"
            )

    lines.append("")
    lines.append("-" * 78)
    lines.append(" Signif. codes: '*' confidence band does not cover 0")
    lines.append("")

    control_group = result.estimation_params.get("control_group")
    if control_group:
        control_text = {"nevertreated": "Never Treated", "notyettreated": "Not Yet Treated"}.get(
            control_group, control_group
        )
        lines.append(f" Control Group: {control_text}")

    anticipation = result.estimation_params.get("anticipation_periods", 0)
    lines.append(f" Anticipation Periods: {anticipation}")

    est_method = result.estimation_params.get("estimation_method")
    if est_method:
        method_text = {"dr": "Doubly Robust", "ipw": "Inverse Probability Weighting", "reg": "Outcome Regression"}.get(
            est_method, est_method
        )
        lines.append(f" Estimation Method: {method_text}")

    lines.append("=" * 78)

    return "\n".join(lines)


def _get_z_critical(alpha: float):
    """Get critical value from standard normal distribution."""
    from scipy import stats

    return stats.norm.ppf(1 - alpha)


def _aggte_repr(self):
    return format_aggte_result(self)


def _aggte_str(self):
    return format_aggte_result(self)


AGGTEResult.__repr__ = _aggte_repr
AGGTEResult.__str__ = _aggte_str



"""Interface for delta method selection."""

from .delta.rm import (
    compute_conditional_cs_rm,
    compute_conditional_cs_rmb,
    compute_conditional_cs_rmm,
)
from .delta.sd import (
    compute_conditional_cs_sd,
    compute_conditional_cs_sdb,
    compute_conditional_cs_sdm,
)
from .delta.sdrm import (
    compute_conditional_cs_sdrm,
    compute_conditional_cs_sdrmb,
    compute_conditional_cs_sdrmm,
)


class DeltaMethodSelector:
    """Select appropriate delta method based on configuration."""

    @staticmethod
    def get_smoothness_method(
        monotonicity_direction,
        bias_direction,
    ):
        """Get delta method for smoothness restrictions.

        Parameters
        ----------
        monotonicity_direction : str, optional
            Direction of monotonicity restriction: "increasing" or "decreasing".
        bias_direction : str, optional
            Direction of bias restriction: "positive" or "negative".

        Returns
        -------
        tuple
            A tuple of (method_function, delta_type_name).
        """
        if monotonicity_direction is None and bias_direction is None:
            return compute_conditional_cs_sd, "DeltaSD"
        if monotonicity_direction is not None:
            delta_type = "DeltaSDI" if monotonicity_direction == "increasing" else "DeltaSDD"
            return compute_conditional_cs_sdm, delta_type
        delta_type = "DeltaSDPB" if bias_direction == "positive" else "DeltaSDNB"
        return compute_conditional_cs_sdb, delta_type

    @staticmethod
    def get_relative_magnitude_method(
        bound_type,
        monotonicity_direction,
        bias_direction,
    ):
        """Get delta method for relative magnitude restrictions.

        Parameters
        ----------
        bound_type : str
            Type of bound: "deviation from parallel trends" or "deviation from linear trend".
        monotonicity_direction : str, optional
            Direction of monotonicity restriction: "increasing" or "decreasing".
        bias_direction : str, optional
            Direction of bias restriction: "positive" or "negative".

        Returns
        -------
        tuple
            A tuple of (method_function, delta_type_name).
        """
        if bound_type == "deviation from parallel trends":
            if monotonicity_direction is None and bias_direction is None:
                return compute_conditional_cs_rm, "DeltaRM"
            if monotonicity_direction is not None:
                delta_type = "DeltaRMI" if monotonicity_direction == "increasing" else "DeltaRMD"
                return compute_conditional_cs_rmm, delta_type
            delta_type = "DeltaRMPB" if bias_direction == "positive" else "DeltaRMNB"
            return compute_conditional_cs_rmb, delta_type
        if monotonicity_direction is None and bias_direction is None:
            return compute_conditional_cs_sdrm, "DeltaSDRM"
        if monotonicity_direction is not None:
            delta_type = "DeltaSDRMI" if monotonicity_direction == "increasing" else "DeltaSDRMD"
            return compute_conditional_cs_sdrmm, delta_type
        delta_type = "DeltaSDRMPB" if bias_direction == "positive" else "DeltaSDRMNB"
        return compute_conditional_cs_sdrmb, delta_type


def get_delta_method(
    sensitivity_type,
    bound_type,
    monotonicity_direction,
    bias_direction,
):
    """Get appropriate delta method based on parameters.

    Parameters
    ----------
    sensitivity_type : {'smoothness', 'relative_magnitude'}
        Type of sensitivity analysis.
    bound_type : str, default="deviation from parallel trends"
        Type of bound for relative magnitude methods.
    monotonicity_direction : str, optional
        Direction of monotonicity restriction: "increasing" or "decreasing".
    bias_direction : str, optional
        Direction of bias restriction: "positive" or "negative".

    Returns
    -------
    tuple
        A tuple of (method_function, delta_type_name).

    Raises
    ------
    ValueError
        If invalid sensitivity_type is provided.
    """
    if sensitivity_type == "smoothness":
        return DeltaMethodSelector.get_smoothness_method(
            monotonicity_direction=monotonicity_direction,
            bias_direction=bias_direction,
        )
    if sensitivity_type == "relative_magnitude":
        return DeltaMethodSelector.get_relative_magnitude_method(
            bound_type=bound_type,
            monotonicity_direction=monotonicity_direction,
            bias_direction=bias_direction,
        )
    raise ValueError(f"sensitivity_type must be 'smoothness' or 'relative_magnitude', got {sensitivity_type}")



"""Datasets."""

import gzip
import pickle
from pathlib import Path

import numpy as np
import pandas as pd

__all__ = ["load_nsw", "load_mpdta", "load_ehec"]


def load_nsw() -> pd.DataFrame:
    """Load the NSW (National Supported Work) demonstration dataset.

    This dataset is from the National Supported Work (NSW) Demonstration,
    a randomized employment training program operated in the mid-1970s.
    It has been widely used in the causal inference literature, particularly
    for demonstrating difference-in-differences methods.

    The dataset is in long format with observations for 1975 (pre-treatment)
    and 1978 (post-treatment) periods.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the following columns:

        - *id*: Individual identifier
        - *year*: Year (1975 or 1978)
        - *experimental*: Treatment indicator (1 if treated, 0 if control)
        - *re*: Real earnings (outcome variable)
        - *age*: Age in years
        - *educ*: Years of education
        - *black*: Indicator for Black race
        - *married*: Indicator for married status
        - *nodegree*: Indicator for no high school degree
        - *hisp*: Indicator for Hispanic ethnicity
        - *re74*: Real earnings in 1974

    Notes
    -----
    This dataset was used in Lalonde (1986) and has been extensively analyzed
    in the treatment effects literature. The version included here is formatted
    for panel data difference-in-differences analysis.

    References
    ----------

    .. [1] Lalonde, R. (1986). Evaluating the econometric evaluations of
        training programs with experimental data. American Economic Review,
        76(4), 604-620.
    """
    data_path = Path(__file__).parent / "_data" / "nsw_long.pkl.gz"

    if not data_path.exists():
        raise FileNotFoundError(
            f"NSW data file not found at {data_path}. "
            "Please ensure the data file is included in the moderndid installation."
        )

    with gzip.open(data_path, "rb") as f:
        nsw_data = pickle.load(f)

    return nsw_data


def load_mpdta() -> pd.DataFrame:
    """Load the County Teen Employment dataset for multiple time period DiD analysis.

    This dataset contains county-level teen employment rates from 2003-2007
    with staggered treatment timing (minimum wage increases). States were first
    treated in 2004, 2006, or 2007.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the following columns:

        - *year*: Year (2003-2007)
        - *countyreal*: County identifier
        - *lpop*: Log of county population
        - *lemp*: Log of county-level teen employment (outcome variable)
        - *first.treat*: Period when state first increased minimum wage (2004, 2006, 2007, or 0 for never-treated)
        - *treat*: Treatment indicator (1 if treated, 0 if control)

    References
    ----------

    .. [1] Callaway, B., & Sant'Anna, P. H. (2021). Difference-in-differences
        with multiple time periods. Journal of Econometrics, 225(2), 200-230.
    """
    data_path = Path(__file__).parent / "_data" / "mpdta_long.pkl.gz"

    if not data_path.exists():
        raise FileNotFoundError(
            f"MPDTA data file not found at {data_path}. "
            "Please ensure the data file is included in the moderndid installation."
        )

    with gzip.open(data_path, "rb") as f:
        mpdta_data = pickle.load(f)

    mpdta_data["first.treat"] = mpdta_data["first.treat"].astype(np.int64)

    return mpdta_data


def load_ehec() -> pd.DataFrame:
    """Load the EHEC dataset for Medicaid expansion analysis.

    This dataset contains state-level data on health insurance coverage rates
    among low-income childless adults from 2008-2019, used to study the effects
    of Medicaid expansion under the Affordable Care Act.

    The dataset tracks states that expanded Medicaid at different times
    (2014, 2015, 2016, 2017, or 2019) as well as states that never expanded
    during the sample period.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the following columns:

        - *stfips*: State FIPS code identifier
        - *year*: Year (2008-2019)
        - *dins*: Share of low-income childless adults with health insurance (outcome variable)
        - *yexp2*: Year that state expanded Medicaid (2014, 2015, 2016, 2017, 2019, or NaN for never-expanded)
        - *W*: State population weights

    Notes
    -----
    This dataset is commonly used in staggered adoption difference-in-differences
    settings and for demonstrating methods that account for treatment effect
    heterogeneity across time and cohorts.

    The data comes from the Mixtape Sessions Advanced DID workshop and is used
    in examples demonstrating the HonestDiD method for sensitivity analysis.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    data_path = Path(__file__).parent / "_data" / "ehec_data.pkl.gz"

    if not data_path.exists():
        raise FileNotFoundError(
            f"EHEC data file not found at {data_path}. "
            "Please ensure the data file is included in the moderndid installation."
        )

    with gzip.open(data_path, "rb") as f:
        ehec_data = pickle.load(f)

    return ehec_data



"""DiD preprocessing."""

from .preprocess.builders import DIDDataBuilder
from .preprocess.constants import BasePeriod, ControlGroup, EstimationMethod
from .preprocess.models import DIDConfig


def preprocess_did(
    data,
    yname,
    tname,
    gname,
    idname=None,
    xformla=None,
    panel=True,
    allow_unbalanced_panel=True,
    control_group="nevertreated",
    anticipation=0,
    weightsname=None,
    alp=0.05,
    bstrap=False,
    cband=False,
    biters=1000,
    clustervars=None,
    est_method="dr",
    base_period="varying",
    print_details=True,
    faster_mode=False,
    pl=False,
    cores=1,
):
    """Process data for multi-period difference-in-differences.

    Parameters
    ----------
    data : pd.DataFrame
        Panel or repeated cross-section data.
    yname : str
        Name of outcome variable column.
    tname : str
        Name of time period column.
    gname : str
        Name of treatment group column. Should contain the time period
        when a unit is first treated (0 for never-treated).
    idname : str | None, default None
        Name of entity/unit identifier column. Required for panel data.
    xformla : str | None, default None
        Formula for covariates in Wilkinson notation (e.g., "~ x1 + x2").
        If None, no covariates are included.
    panel : bool, default True
        Whether data is in panel format (vs repeated cross-sections).
    allow_unbalanced_panel : bool, default True
        Whether to allow unbalanced panels.
    control_group : {"nevertreated", "notyettreated"}, default "nevertreated"
        Which units to use as control group.
    anticipation : int, default 0
        Number of time periods before treatment where effects may appear.
    weightsname : str | None, default None
        Name of sampling weights column.
    alp : float, default 0.05
        Significance level for confidence intervals.
    bstrap : bool, default False
        Whether to use bootstrap for inference.
    cband : bool, default False
        Whether to compute uniform confidence bands.
    biters : int, default 1000
        Number of bootstrap iterations.
    clustervars : list[str] | None, default None
        Variables to cluster standard errors on.
    est_method : {"dr", "ipw", "reg"}, default "dr"
        Estimation method: doubly robust, IPW, or regression.
    base_period : {"universal", "varying"}, default "varying"
        How to choose base period for comparisons.
    print_details : bool, default True
        Whether to print progress messages.
    faster_mode : bool, default False
        Whether to use computational shortcuts.
    pl : bool, default False
        Whether to use parallel processing.
    cores : int, default 1
        Number of cores for parallel processing.

    Returns
    -------
    DIDData
        Container with all preprocessed data and parameters including:

        - data: Standardized panel/cross-section data
        - weights: Normalized sampling weights
        - config: Configuration with all settings
        - Various tensors and matrices for computation
    """
    control_group_enum = ControlGroup(control_group)
    est_method_enum = EstimationMethod(est_method)
    base_period_enum = BasePeriod(base_period)

    config = DIDConfig(
        yname=yname,
        tname=tname,
        idname=idname,
        gname=gname,
        xformla=xformla if xformla is not None else "~1",
        panel=panel,
        allow_unbalanced_panel=allow_unbalanced_panel,
        control_group=control_group_enum,
        anticipation=anticipation,
        weightsname=weightsname,
        alp=alp,
        bstrap=bstrap,
        cband=cband,
        biters=biters,
        clustervars=clustervars if clustervars is not None else [],
        est_method=est_method_enum,
        base_period=base_period_enum,
        print_details=print_details,
        faster_mode=faster_mode,
        pl=pl,
        cores=cores,
    )

    builder = DIDDataBuilder()
    did_data = builder.with_data(data).with_config(config).validate().transform().build()

    return did_data



"""Tensor creation factories for DiD preprocessing."""

from abc import ABC, abstractmethod
from typing import Protocol

import numpy as np
import pandas as pd

from moderndid.utils import extract_vars_from_formula

from .constants import WEIGHTS_COLUMN, DataFormat
from .models import DIDConfig


class TensorFactory(Protocol):
    """Protocol for tensor factories."""

    def create_tensors(self, data: pd.DataFrame, config: DIDConfig) -> dict[str, np.ndarray | list[np.ndarray] | None]:
        """Create tensors from data."""


class BaseTensorFactory(ABC):
    """Base class for tensor factories."""

    @abstractmethod
    def create_outcomes_tensor(self, data: pd.DataFrame, config: DIDConfig) -> list[np.ndarray] | None:
        """Create outcomes tensor."""

    @abstractmethod
    def create_covariates_tensor(self, data: pd.DataFrame, config: DIDConfig) -> list[np.ndarray] | np.ndarray | None:
        """Create covariates tensor or matrix."""

    @staticmethod
    def create_time_invariant_data(data: pd.DataFrame, config: DIDConfig) -> pd.DataFrame:
        """Extract time-invariant data (one row per unit)."""
        time_invariant_cols = [config.idname, config.gname, WEIGHTS_COLUMN]

        if config.clustervars:
            time_invariant_cols.extend(config.clustervars)

        # Add time-invariant covariates if any
        if config.xformla and config.xformla != "~1":
            formula_vars = extract_vars_from_formula(config.xformla)
            formula_vars = [v for v in formula_vars if v != config.yname]

            # Check which variables are time-invariant
            for var in formula_vars:
                if var in data.columns:
                    var_counts = data.groupby(config.idname)[var].nunique()
                    if (var_counts == 1).all():
                        time_invariant_cols.append(var)

        time_invariant_cols = list(dict.fromkeys(time_invariant_cols))

        return (
            data.groupby(config.idname)
            .first()[[col for col in time_invariant_cols if col != config.idname]]
            .reset_index()
        )

    @staticmethod
    def create_summary_tables(
        data: pd.DataFrame, time_invariant_data: pd.DataFrame, config: DIDConfig
    ) -> dict[str, pd.DataFrame]:
        """Create summary tables for cohorts and periods."""
        # Cohort counts
        cohort_counts = time_invariant_data.groupby(config.gname).size().reset_index(name="cohort_size")
        cohort_counts.columns = ["cohort", "cohort_size"]

        # Period counts
        period_counts = data.groupby(config.tname).size().reset_index(name="period_size")
        period_counts.columns = ["period", "period_size"]

        # Crosstable counts
        crosstable = data.groupby([config.tname, config.gname]).size().reset_index(name="count")
        crosstable.columns = ["period", "cohort", "count"]
        crosstable_counts = crosstable.pivot(index="period", columns="cohort", values="count").fillna(0)

        return {
            "cohort_counts": cohort_counts,
            "period_counts": period_counts,
            "crosstable_counts": crosstable_counts,
        }

    @staticmethod
    def extract_cluster_variable(time_invariant_data: pd.DataFrame, config: DIDConfig) -> np.ndarray | None:
        """Extract cluster variable if specified."""
        if config.clustervars and len(config.clustervars) > 0:
            return time_invariant_data[config.clustervars[0]].values
        return None

    @staticmethod
    def extract_weights(time_invariant_data: pd.DataFrame) -> np.ndarray:
        """Extract normalized weights."""
        return time_invariant_data[WEIGHTS_COLUMN].values


class PanelTensorFactory(BaseTensorFactory):
    """Factory for balanced panel data tensors."""

    def create_outcomes_tensor(self, data: pd.DataFrame, config: DIDConfig) -> list[np.ndarray]:
        """Create list of outcome arrays, one per time period."""
        outcomes_tensor = []

        for i in range(len(config.time_periods)):
            start_idx = i * config.id_count
            end_idx = (i + 1) * config.id_count
            outcomes_tensor.append(data[config.yname].iloc[start_idx:end_idx].values)

        return outcomes_tensor

    def create_covariates_tensor(self, data: pd.DataFrame, config: DIDConfig) -> list[np.ndarray]:
        """Create list of covariate matrices, one per time period."""
        if config.xformla == "~1" or config.xformla is None:
            return [np.ones((config.id_count, 1)) for _ in range(config.time_periods_count)]

        formula_vars = extract_vars_from_formula(config.xformla)
        formula_vars = [v for v in formula_vars if v != config.yname]

        covariates_tensor = []
        for i in range(config.time_periods_count):
            start_idx = i * config.id_count
            end_idx = (i + 1) * config.id_count
            period_data = data.iloc[start_idx:end_idx]

            X = np.column_stack([np.ones(len(period_data))] + [period_data[v].values for v in formula_vars])
            covariates_tensor.append(X)

        return covariates_tensor


class UnbalancedPanelTensorFactory(BaseTensorFactory):
    """Factory for unbalanced panel data."""

    def create_outcomes_tensor(self, data: pd.DataFrame, config: DIDConfig) -> None:
        """No outcomes tensor for unbalanced panels."""
        return None

    def create_covariates_tensor(self, data: pd.DataFrame, config: DIDConfig) -> np.ndarray:
        """Create single covariate matrix using time-invariant data."""
        time_invariant_data = self.__class__.create_time_invariant_data(data, config)

        if config.xformla == "~1" or config.xformla is None:
            return np.ones((len(time_invariant_data), 1))

        formula_vars = extract_vars_from_formula(config.xformla)
        formula_vars = [v for v in formula_vars if v != config.yname]

        # Filter to only time-invariant variables
        available_vars = []
        for var in formula_vars:
            if var in time_invariant_data.columns:
                available_vars.append(var)
            else:
                # Try to get from original data if truly time-invariant
                var_counts = data.groupby(config.idname)[var].nunique()
                if (var_counts == 1).all():
                    time_invariant_data[var] = data.groupby(config.idname)[var].first().values
                    available_vars.append(var)

        X = np.column_stack(
            [np.ones(len(time_invariant_data))] + [time_invariant_data[v].values for v in available_vars]
        )

        return X


class RepeatedCrossSectionTensorFactory(BaseTensorFactory):
    """Factory for repeated cross-section data."""

    def create_outcomes_tensor(self, data: pd.DataFrame, config: DIDConfig) -> None:
        """No outcomes tensor for repeated cross-sections."""
        return None

    def create_covariates_tensor(self, data: pd.DataFrame, config: DIDConfig) -> np.ndarray:
        """Create covariate matrix from full data."""
        if config.xformla == "~1" or config.xformla is None:
            return np.ones((len(data), 1))

        formula_vars = extract_vars_from_formula(config.xformla)
        formula_vars = [v for v in formula_vars if v != config.yname]

        X = np.column_stack([np.ones(len(data))] + [data[v].values for v in formula_vars])

        return X


class TensorFactorySelector:
    """Selects appropriate tensor factory based on data format."""

    @staticmethod
    def get_factory(config: DIDConfig) -> BaseTensorFactory:
        """Get appropriate tensor factory."""
        if config.data_format == DataFormat.PANEL:
            return PanelTensorFactory()
        if config.data_format == DataFormat.UNBALANCED_PANEL:
            return UnbalancedPanelTensorFactory()
        if config.data_format == DataFormat.REPEATED_CROSS_SECTION:
            return RepeatedCrossSectionTensorFactory()
        raise ValueError(f"Unknown data format: {config.data_format}")

    @classmethod
    def create_tensors(
        cls, data: pd.DataFrame, config: DIDConfig
    ) -> dict[str, np.ndarray | list[np.ndarray] | pd.DataFrame | None]:
        """Create all tensors using appropriate factory."""
        factory = cls.get_factory(config)

        time_invariant_data = factory.create_time_invariant_data(data, config)
        summary_tables = factory.create_summary_tables(data, time_invariant_data, config)

        outcomes_tensor = factory.create_outcomes_tensor(data, config)
        covariates_tensor = factory.create_covariates_tensor(data, config)

        cluster = factory.extract_cluster_variable(time_invariant_data, config)
        weights = factory.extract_weights(time_invariant_data)

        if isinstance(covariates_tensor, list):
            covariates_matrix = None
        else:
            covariates_matrix = covariates_tensor
            covariates_tensor = None

        return {
            "data": data,
            "time_invariant_data": time_invariant_data,
            "outcomes_tensor": outcomes_tensor,
            "covariates_matrix": covariates_matrix,
            "covariates_tensor": covariates_tensor,
            "cluster": cluster,
            "weights": weights,
            **summary_tables,
        }



"""Difference-in-Differences estimators with multiple time periods."""

from .aggte import aggte
from .aggte_obj import AGGTEResult, format_aggte_result
from .att_gt import att_gt
from .compute_aggte import compute_aggte
from .compute_att_gt import ATTgtResult, ComputeATTgtResult, compute_att_gt
from .mboot import mboot
from .multiperiod_obj import (
    MPPretestResult,
    MPResult,
    format_mp_pretest_result,
    format_mp_result,
    mp,
    mp_pretest,
    summary_mp_pretest,
)
from .plots import (
    plot_att_gt,
    plot_did,
    plot_event_study,
)
from .preprocess.models import DIDData
from .preprocess_did import preprocess_did

__all__ = [
    "aggte",
    "AGGTEResult",
    "format_aggte_result",
    "compute_aggte",
    "att_gt",
    "MPResult",
    "mp",
    "format_mp_result",
    "MPPretestResult",
    "mp_pretest",
    "format_mp_pretest_result",
    "summary_mp_pretest",
    "preprocess_did",
    "DIDData",
    "mboot",
    "ATTgtResult",
    "ComputeATTgtResult",
    "compute_att_gt",
    "plot_att_gt",
    "plot_event_study",
    "plot_did",
]
